#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from .pressure_far_field_child import pressure_far_field_child

class pressure_far_field(NamedObject[pressure_far_field_child], _NonCreatableNamedObjectMixin[pressure_far_field_child]):
    """
    'pressure_far_field' child.
    """

    fluent_name = "pressure-far-field"

    child_object_type: pressure_far_field_child = pressure_far_field_child
    """
    child_object_type of pressure_far_field.
    """



import pytest
from util.solver_workflow import new_solver_session  # noqa: F401

from ansys.fluent.core.services.datamodel_tui import TUIMenu


@pytest.mark.skip("randomly failing due to missing transcript capture")
def test_report_system_proc_stats_tui(new_solver_session, capsys) -> None:
    new_solver_session.transcript.start()
    # Issue: Transcript missing for the first TUI command
    new_solver_session.solver.tui.report.system.sys_stats()
    new_solver_session.solver.tui.report.system.sys_stats()
    captured = capsys.readouterr()
    assert "CPU" in captured.out


def test_runtime_tui_menus(load_static_mixer_case) -> None:
    solver = load_static_mixer_case
    solver.tui.define.models.addon_module(3)
    rmf = solver.tui.define.models.resolved_MEA_fuelcells
    assert rmf is not None
    assert rmf.__class__ == TUIMenu



from util.meshing_workflow import new_mesh_session  # noqa: F401
from util.solver_workflow import new_solver_session  # noqa: F401


def test_solver_preferences(new_solver_session):
    solver = new_solver_session

    solver.preferences.MeshingWorkflow.Verbosity = "off"
    assert solver.preferences.MeshingWorkflow.Verbosity() == "off"

    solver.preferences.MeshingWorkflow.CheckpointingOption = "Write into memory"
    assert (
        solver.preferences.MeshingWorkflow.CheckpointingOption() == "Write into memory"
    )

    solver.preferences.MeshingWorkflow.DrawSettings.FacetLimit = 6000000
    assert solver.preferences.MeshingWorkflow.DrawSettings.FacetLimit() == 6000000

    solver.preferences.MeshingWorkflow.DrawSettings.FaceZoneLimit = 15000
    solver.preferences.MeshingWorkflow.DrawSettings.FaceZoneLimit() == 15000

    solver.preferences.Appearance.AnsysLogo.Color = "white"
    assert solver.preferences.Appearance.AnsysLogo.Color() == "white"

    solver.preferences.Appearance.AnsysLogo.Color = "black"
    assert solver.preferences.Appearance.AnsysLogo.Color() == "black"

    solver.preferences.Appearance.AnsysLogo.Visible = True
    assert solver.preferences.Appearance.AnsysLogo.Visible() is True

    solver.preferences.Graphics.AnimationOption = "wireframe"
    assert solver.preferences.Graphics.AnimationOption() == "wireframe"

    solver.exit()


def test_meshing_preferences(new_mesh_session):
    meshing = new_mesh_session

    meshing.preferences.MeshingWorkflow.Verbosity = "off"
    assert meshing.preferences.MeshingWorkflow.Verbosity() == "off"

    meshing.preferences.MeshingWorkflow.CheckpointingOption = "Write into memory"
    assert (
        meshing.preferences.MeshingWorkflow.CheckpointingOption() == "Write into memory"
    )

    meshing.preferences.MeshingWorkflow.DrawSettings.FacetLimit = 6000000
    assert meshing.preferences.MeshingWorkflow.DrawSettings.FacetLimit() == 6000000

    meshing.preferences.MeshingWorkflow.DrawSettings.FaceZoneLimit = 15000
    meshing.preferences.MeshingWorkflow.DrawSettings.FaceZoneLimit() == 15000

    meshing.preferences.Appearance.AnsysLogo.Color = "white"
    assert meshing.preferences.Appearance.AnsysLogo.Color() == "white"

    meshing.preferences.Appearance.AnsysLogo.Color = "black"
    assert meshing.preferences.Appearance.AnsysLogo.Color() == "black"

    meshing.preferences.Appearance.AnsysLogo.Visible = True
    assert meshing.preferences.Appearance.AnsysLogo.Visible() is True

    meshing.preferences.Graphics.AnimationOption = "wireframe"
    assert meshing.preferences.Graphics.AnimationOption() == "wireframe"

    meshing.exit()



import logging
from typing import Iterator, Tuple

from ansys.fluent.core.services.datamodel_se import PyCallableStateObject

datamodel_logger = logging.getLogger("ansys.fluent.services.datamodel")


def _new_command_for_task(task, session):
    class NewCommandError(Exception):
        def __init__(self, task_name):
            super().__init__(f"Could not create command for meshing task {task_name}")

    task_cmd_name = task.CommandName()
    cmd_creator = getattr(session, task_cmd_name)
    if cmd_creator:
        new_cmd = cmd_creator.create_instance()
        if new_cmd:
            return new_cmd
    raise NewCommandError(task._name_())


class Task(PyCallableStateObject):
    """Wrap a Workflow TaskObject instance, adding methods to discover more
    about the relationships between TaskObjects.

    Methods
    -------
    get_direct_upstream_tasks()
    get_direct_downstream_tasks()
    ordered_children()
    inactive_ordered_children()
    get_id()
    get_idx()
    __getattr__(attr)
    __setattr__(attr, value)
    __dir__()
    """

    def __init__(self, command_source, name: str) -> None:
        self.__dict__.update(
            dict(
                _command_source=command_source,
                _workflow=command_source._workflow,
                _source=command_source._command_source,
                _task=command_source._workflow.TaskObject[name],
                _cmd=None,
            )
        )

    def get_direct_upstream_tasks(self) -> list:
        """Get the list of tasks upstream of this one and directly connected by
        a data dependency.

        Returns
        -------
        upstreams : list
            Upstream task list.
        """
        return self._tasks_with_matching_attributes(
            attr="requiredInputs", other_attr="outputs"
        )

    def get_direct_downstream_tasks(self) -> list:
        """Get the list of tasks downstream of this one and directly connected
        by a data dependency.

        Returns
        -------
        downstreams : list
            Downstream task list.
        """
        return self._tasks_with_matching_attributes(
            attr="outputs", other_attr="requiredInputs"
        )

    def ordered_children(self) -> list:
        """Get the ordered task list held by this task. Sorting is in terms
        of the workflow order and only includes this task's top-level tasks, while other tasks
        can be obtained by calling ordered_children() on a parent task. Given the
        workflow::

            Workflow
            ├── A
            ├── B
            │   ├── C
            │   └── D
            └── E

        C and D are the ordered children of task B.

        Returns
        -------
        children : list
            Ordered children.
        """
        return [
            self._command_source._task_by_id(task_id)
            for task_id in self._task.TaskList()
        ]

    def inactive_ordered_children(self) -> list:
        """Get the inactive ordered task list held by this task.

        Returns
        -------
        children : list
            Inactive ordered children.
        """
        return [
            self._command_source._task_by_id(task_id)
            for task_id in self._task.InactiveTaskList()
        ]

    def get_id(self) -> str:
        """Get the unique string identifier of this task, as it is in the
        meshing application.

        Returns
        -------
        identifier : str
            The string identifier.
        """
        workflow_state = self._command_source._workflow_state()
        for k, v in workflow_state.items():
            if isinstance(v, dict) and "_name_" in v:
                if v["_name_"] == self.name():
                    type_, id_ = k.split(":")
                    if type_ == "TaskObject":
                        return id_

    def get_idx(self) -> int:
        """Get the unique integer index of this task, as it is in the meshing
        application.

        Returns
        -------
        index : int
            The integer index.
        """
        return int(self.get_id()[len("TaskObject") :])

    @property
    def CommandArguments(self):
        return self._refreshed_command()

    def __getattr__(self, attr):
        return getattr(self._task, attr)

    def __setattr__(self, attr, value):
        if attr in self.__dict__:
            self.__dict__[attr] = value
            datamodel_logger.debug(f"Set {attr}  to {value}")
        else:
            setattr(self._task, attr, value)
            datamodel_logger.debug(f"Set {attr}  to {value}")

    def __dir__(self):
        return sorted(
            set(list(self.__dict__.keys()) + dir(type(self)) + dir(self._task))
        )

    def _tasks_with_matching_attributes(self, attr: str, other_attr: str) -> list:
        this_command = self._command()
        attrs = this_command.get_attr(attr)
        if not attrs:
            return []
        attrs = set(attrs)
        tasks = [
            task
            for task in self._command_source.ordered_children()
            if task.name() != self.name()
        ]
        matches = []
        for task in tasks:
            command = task._command()
            other_attrs = command.get_attr(other_attr)
            if other_attrs and (attrs & set(other_attrs)):
                matches.append(task)
        return matches

    def _refreshed_command(self):
        task_arg_state = self.Arguments.get_state()
        cmd = self._command()
        if task_arg_state:
            cmd.set_state(task_arg_state)
        return _MakeReadOnly(self._cmd_sub_items_read_only(cmd))

    def _cmd_sub_items_read_only(self, cmd):
        for item in cmd():
            if type(getattr(cmd, item).get_state()) == dict:
                setattr(cmd, item, self._cmd_sub_items_read_only(getattr(cmd, item)))
            setattr(cmd, item, _MakeReadOnly(getattr(cmd, item)))
        return cmd

    def _command(self):
        if not self._cmd:
            self._cmd = _new_command_for_task(self._task, self._source)
        return self._cmd


class TaskContainer(PyCallableStateObject):
    """Wrap a workflow TaskObject container.

    Methods
    -------
    __iter__()
    __getitem__(attr)
    __getattr__(attr)
    __dir__()
    """

    def __init__(self, command_source):
        self._container = command_source
        self._task_container = command_source._workflow.TaskObject

    def __iter__(self) -> Iterator[Task]:
        """Yield the next child object.

        Yields
        ------
        Iterator[Task]
            Iterator of child objects.
        """
        for name in self._get_child_object_display_names():
            yield self[name]

    def __getitem__(self, name):
        datamodel_logger.debug(f"Task: {name}")
        return Task(self._container, name)

    def __getattr__(self, attr):
        return getattr(self._task_container, attr)

    def __dir__(self):
        return sorted(
            set(
                list(self.__dict__.keys()) + dir(type(self)) + dir(self._task_container)
            )
        )


class WorkflowWrapper:
    """Wrap a Workflow object, adding methods to discover more about the
    relationships between TaskObjects.

    Methods
    -------
    task(name)
    ordered_children()
    __getattr__(attr)
    __dir__()
    __call__()
    """

    def __init__(self, workflow, command_source):
        self._workflow = workflow
        self._command_source = command_source

    def task(self, name: str) -> Task:
        """Get a TaskObject by name, in a Task wrapper. The wrapper adds extra
        functionality.

        Parameters
        ----------
        name : str
            Task name - the display name, not the internal ID.

        Returns
        -------
        task : Task
            wrapped task object.
        """
        return Task(self, name)

    @property
    def TaskObject(self) -> TaskContainer:
        # missing from dir
        """Get a TaskObject container wrapper that 'holds' the underlying
        TaskObjects.

        The wrapper adds extra functionality.
        """
        return TaskContainer(self)

    def ordered_children(self) -> list:
        """Get the ordered task list held by the workflow. Sorting is in terms
        of the workflow order and only includes the top-level tasks, while other tasks
        can be obtained by calling ordered_children() on a parent task. Given the
        workflow::

            Workflow
            ├── A
            ├── B
            │   ├── C
            │   └── D
            └── E

        the ordered children of the workflow are A, B, E, while B has ordered children
        C and D.
        """
        workflow_state, task_list_state = self._workflow_and_task_list_state()
        tasks = []
        for task_id in task_list_state:
            tasks.append(self._task_by_id_impl(task_id, workflow_state))
        return tasks

    def __getattr__(self, attr):
        """Delegate attribute lookup to the wrapped workflow object
        Parameters
        ----------
        attr : str
            An attribute not defined in WorkflowWrapper
        """
        return getattr(self._workflow, attr)

    def __dir__(self):
        """Override the behaviour of dir to include attributes in
        WorkflowWrapper and the underlying workflow."""
        return sorted(
            set(list(self.__dict__.keys()) + dir(type(self)) + dir(self._workflow))
        )

    def __call__(self):
        """Delegate calls to the underlying workflow."""
        return self._workflow()

    def _workflow_state(self):
        return self._workflow()

    def _workflow_and_task_list_state(self) -> Tuple[dict, dict]:
        workflow_state = self._workflow_state()
        workflow_state_workflow = workflow_state["Workflow"]
        return (workflow_state, workflow_state_workflow["TaskList"])

    def _task_by_id_impl(self, task_id, workflow_state):
        task_key = "TaskObject:" + task_id
        task_state = workflow_state[task_key]
        return self.task(task_state["_name_"])

    def _task_by_id(self, task_id):
        workflow_state = self._workflow_state()
        return self._task_by_id_impl(task_id, workflow_state)


class _MakeReadOnly:
    """Removes 'set_state()' attribute to implement read-only behaviour."""

    _unwanted_attr = ["set_state", "setState"]

    def __init__(self, cmd):
        self._cmd = cmd

    def is_read_only(self):
        return True

    def __getattr__(self, attr):
        if attr in _MakeReadOnly._unwanted_attr:
            raise AttributeError("Command Arguments are read-only.")
        return getattr(self._cmd, attr)

    def __dir__(self):
        returned_list = sorted(
            set(list(self.__dict__.keys()) + dir(type(self)) + dir(self._cmd))
        )
        for attr in _MakeReadOnly._unwanted_attr:
            if attr in returned_list:
                returned_list.remove(attr)
        return returned_list

    def __call__(self):
        return self._cmd()



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class inj_mass_rate_last_tstp(Real):
    """
    'inj_mass_rate_last_tstp' child.
    """

    fluent_name = "inj-mass-rate-last-tstp"




import pytest
from util.solver_workflow import new_solver_session  # noqa: F401

from ansys.fluent.core.examples import download_file


@pytest.mark.dev
@pytest.mark.fluent_231
@pytest.mark.fluent_232
def test_setup_models_viscous_model_settings(new_solver_session) -> None:
    solver_session = new_solver_session
    case_path = download_file("elbow_source_terms.cas.h5", "pyfluent/mixing_elbow")
    solver_session.file.read_case(file_name=case_path)
    solver_session.solution.initialization.hybrid_initialize()
    assert solver_session.setup.models.viscous.model() == "k-epsilon"
    assert "inviscid" in solver_session.setup.models.viscous.model.get_attr(
        "allowed-values"
    )
    solver_session.setup.models.viscous.model = "inviscid"
    assert solver_session.setup.models.viscous.model() == "inviscid"


@pytest.mark.skip(reason="failing if run with other tests")
def test_results_graphics_mesh_settings(new_solver_session) -> None:
    session = new_solver_session
    case_path = download_file("Static_Mixer_main.cas.h5", "pyfluent/static_mixer")
    session.solver.file.read(file_type="case", file_name=case_path)
    assert "mesh-1" not in session.solver.results.graphics.mesh.get_object_names()
    session.solver.results.graphics.mesh["mesh-1"] = {}
    assert "mesh-1" in session.solver.results.graphics.mesh.get_object_names()
    assert not session.solver.results.graphics.mesh["mesh-1"].options.nodes()
    session.solver.results.graphics.mesh["mesh-1"].options.nodes = True
    assert session.solver.results.graphics.mesh["mesh-1"].options.nodes()
    session.solver.results.graphics.mesh.rename("mesh-a", "mesh-1")
    assert "mesh-a" in session.solver.results.graphics.mesh.get_object_names()
    assert "mesh-1" not in session.solver.results.graphics.mesh.get_object_names()
    del session.solver.results.graphics.mesh["mesh-a"]
    assert "mesh-a" not in session.solver.results.graphics.mesh.get_object_names()


@pytest.mark.dev
@pytest.mark.fluent_232
def test_wildcard(new_solver_session):
    solver = new_solver_session
    case_path = download_file("elbow_source_terms.cas.h5", "pyfluent/mixing_elbow")
    solver.file.read_case(file_name=case_path)
    assert solver.setup.boundary_conditions.velocity_inlet["inl*"].vmag() == {
        "inlet2": {"vmag": {"option": "value", "value": 15}},
        "inlet1": {"vmag": {"option": "value", "value": 5}},
    }
    assert solver.setup.boundary_conditions.velocity_inlet["inl*"].vmag.value() == {
        "inlet2": {"vmag": {"value": 15}},
        "inlet1": {"vmag": {"value": 5}},
    }
    solver.setup.boundary_conditions.velocity_inlet["inl*"].vmag = 10
    assert solver.setup.boundary_conditions.velocity_inlet["inl*"].vmag() == {
        "inlet2": {"vmag": {"option": "value", "value": 10}},
        "inlet1": {"vmag": {"option": "value", "value": 10}},
    }
    solver.setup.boundary_conditions.velocity_inlet = (
        solver.setup.boundary_conditions.velocity_inlet["inl*"].vmag()
    )
    assert solver.setup.boundary_conditions.velocity_inlet["inl*"].vmag() == {
        "inlet2": {"vmag": {"option": "value", "value": 10}},
        "inlet1": {"vmag": {"option": "value", "value": 10}},
    }
    assert solver.setup.cell_zone_conditions.fluid["*"].source_terms["*mom*"]() == {
        "fluid": {
            "source_terms": {
                "y-momentum": [{"option": "value", "value": 2}],
                "x-momentum": [{"option": "value", "value": 1}],
                "z-momentum": [{"option": "value", "value": 3}],
            }
        }
    }
    solver.setup.cell_zone_conditions.fluid["*"].source_terms["*mom*"] = [
        {"option": "value", "value": 2}
    ]
    assert solver.setup.cell_zone_conditions.fluid["*"].source_terms["*mom*"]() == {
        "fluid": {
            "source_terms": {
                "x-momentum": [{"option": "value", "value": 2}],
                "y-momentum": [{"option": "value", "value": 2}],
                "z-momentum": [{"option": "value", "value": 2}],
            }
        }
    }


@pytest.mark.dev
@pytest.mark.fluent_232
def test_wildcard_fnmatch(new_solver_session):
    solver = new_solver_session
    case_path = download_file("elbow_source_terms.cas.h5", "pyfluent/mixing_elbow")
    solver.file.read_case(file_name=case_path)

    solver.solution.initialization.hybrid_initialize()

    solver.results.graphics.mesh.create("mesh-2")
    solver.results.graphics.mesh.create("mesh-a")
    solver.results.graphics.mesh.create("mesh-bc")

    assert (
        list(solver.results.graphics.mesh["mesh-*"]().keys()).sort()
        == ["mesh-1", "mesh-2", "mesh-a", "mesh-bc"].sort()
    )

    assert list(solver.results.graphics.mesh["mesh-?c"]().keys()) == ["mesh-bc"]

    assert list(solver.results.graphics.mesh["mesh-[2-5]"]().keys()) == ["mesh-2"]

    assert (
        list(solver.results.graphics.mesh["mesh-[!2-5]"]().keys()).sort()
        == ["mesh-1", "mesh-a"].sort()
    )


@pytest.mark.dev
@pytest.mark.fluent_232
def test_wildcard_path_is_iterable(new_solver_session):
    solver = new_solver_session
    case_path = download_file("elbow_source_terms.cas.h5", "pyfluent/mixing_elbow")
    solver.file.read_case(file_name=case_path)

    assert [x for x in solver.setup.boundary_conditions.velocity_inlet] == [
        "inlet2",
        "inlet1",
    ]

    assert [x for x in solver.setup.boundary_conditions.velocity_inlet["*let*"]] == [
        "inlet2",
        "inlet1",
    ]

    assert [x for x in solver.setup.boundary_conditions.velocity_inlet["*1*"]] == [
        "inlet1"
    ]

    test_data = []
    for k, v in solver.setup.boundary_conditions.velocity_inlet.items():
        test_data.append((k, v))

    assert test_data[0][0] == "inlet2"
    assert test_data[0][1].path == r"setup/boundary-conditions/velocity-inlet/inlet2"
    assert test_data[1][0] == "inlet1"
    assert test_data[1][1].path == r"setup/boundary-conditions/velocity-inlet/inlet1"

    test_data = []
    for k, v in solver.setup.boundary_conditions.velocity_inlet["*let*"].items():
        test_data.append((k, v))

    assert test_data[0][0] == "inlet2"
    assert test_data[0][1].path == r"setup/boundary-conditions/velocity-inlet/inlet2"
    assert test_data[1][0] == "inlet1"
    assert test_data[1][1].path == r"setup/boundary-conditions/velocity-inlet/inlet1"



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class max_iter_per_time_step(Integer):
    """
    Set Max Iterations/Time Step.
    """

    fluent_name = "max-iter-per-time-step"




#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class old_default_of_operating_density_method(Boolean):
    """
    'old_default_of_operating_density_method' child.
    """

    fluent_name = "old-default-of-operating-density-method"




#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class fixed_time_step_count(Integer):
    """
     Set Number of Fixed Time Steps.
    """

    fluent_name = "fixed-time-step-count"




#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class compute_from_view_plane(Boolean):
    """
    'compute_from_view_plane' child.
    """

    fluent_name = "compute-from-view-plane?"




#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class enable_dynamic_strength(Boolean):
    """
    Enable dynamic strength to reduce compression in the tangential direction to the interface.
    """

    fluent_name = "enable-dynamic-strength?"




#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class solve_exp_vof_at_end(Boolean):
    """
    Solve Explicit VOF at the end of time-step as an alternative.
    """

    fluent_name = "solve-exp-vof-at-end?"




from concurrent import futures
import os
from pathlib import Path
import tempfile
import time

import grpc
import pytest
from util.meshing_workflow import new_mesh_session  # noqa: F401
from util.solver_workflow import new_solver_session  # noqa: F401

from ansys.api.fluent.v0 import (
    health_pb2,
    health_pb2_grpc,
    scheme_eval_pb2,
    scheme_eval_pb2_grpc,
)
import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples, launch_fluent
from ansys.fluent.core.examples import download_file
from ansys.fluent.core.fluent_connection import FluentConnection
from ansys.fluent.core.session import BaseSession
from ansys.fluent.core.utils.networking import get_free_port


class MockHealthServicer(health_pb2_grpc.HealthServicer):
    def Check(self, request, context: grpc.ServicerContext):  # noqa N802
        metadata = dict(context.invocation_metadata())
        password = metadata.get("password", None)
        if password != "12345":
            context.set_code(grpc.StatusCode.UNAUTHENTICATED)
            return health_pb2.HealthCheckResponse()
        return health_pb2.HealthCheckResponse(
            status=health_pb2.HealthCheckResponse.ServingStatus.SERVING
        )

    def Watch(self, request, context: grpc.ServicerContext):  # noqa N802
        metadata = dict(context.invocation_metadata())
        password = metadata.get("password", None)
        if password != "12345":
            context.set_code(grpc.StatusCode.UNAUTHENTICATED)
            yield health_pb2.HealthCheckResponse()

        c = 0
        while c < 2:
            time.sleep(1)
            c += 1
            yield health_pb2.HealthCheckResponse(
                status=health_pb2.HealthCheckResponse.ServingStatus.NOT_SERVING
            )

        time.sleep(1)
        yield health_pb2.HealthCheckResponse(
            status=health_pb2.HealthCheckResponse.ServingStatus.SERVING
        )


class MockSchemeEvalServicer(scheme_eval_pb2_grpc.SchemeEvalServicer):
    def StringEval(self, request, context):
        if request.input == "(cx-version)":
            return scheme_eval_pb2.StringEvalResponse(output="(23 1 0)")

    def SchemeEval(
        self,
        request,
        context: grpc.ServicerContext,
    ) -> scheme_eval_pb2.SchemeEvalResponse:
        metadata = dict(context.invocation_metadata())
        password = metadata.get("password", None)
        if password != "12345":
            context.set_code(grpc.StatusCode.UNAUTHENTICATED)
            return scheme_eval_pb2.SchemeEvalResponse()


def test_create_session_by_passing_ip_and_port_and_password() -> None:
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
    ip = "127.0.0.1"
    port = get_free_port()
    server.add_insecure_port(f"{ip}:{port}")
    health_pb2_grpc.add_HealthServicer_to_server(MockHealthServicer(), server)
    scheme_eval_pb2_grpc.add_SchemeEvalServicer_to_server(
        MockSchemeEvalServicer(), server
    )
    server.start()
    session = BaseSession(
        FluentConnection(ip=ip, port=port, password="12345", cleanup_on_exit=False)
    )
    assert session.health_check_service.is_serving
    server.stop(None)
    session.exit()
    assert not session.health_check_service.is_serving


def test_create_session_by_setting_ip_and_port_env_var(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
    ip = "127.0.0.1"
    port = get_free_port()
    server.add_insecure_port(f"{ip}:{port}")
    health_pb2_grpc.add_HealthServicer_to_server(MockHealthServicer(), server)
    scheme_eval_pb2_grpc.add_SchemeEvalServicer_to_server(
        MockSchemeEvalServicer(), server
    )
    server.start()
    monkeypatch.setenv("PYFLUENT_FLUENT_IP", ip)
    monkeypatch.setenv("PYFLUENT_FLUENT_PORT", str(port))
    session = BaseSession(FluentConnection(password="12345", cleanup_on_exit=False))
    assert session.health_check_service.is_serving
    server.stop(None)
    session.exit()
    assert not session.health_check_service.is_serving


def test_create_session_by_passing_grpc_channel() -> None:
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
    ip = "127.0.0.1"
    port = get_free_port()
    server.add_insecure_port(f"{ip}:{port}")
    health_pb2_grpc.add_HealthServicer_to_server(MockHealthServicer(), server)
    scheme_eval_pb2_grpc.add_SchemeEvalServicer_to_server(
        MockSchemeEvalServicer(), server
    )
    server.start()
    channel = grpc.insecure_channel(f"{ip}:{port}")
    session = BaseSession(
        FluentConnection(channel=channel, cleanup_on_exit=False, password="12345")
    )
    assert session.health_check_service.is_serving
    server.stop(None)
    session.exit()
    assert not session.health_check_service.is_serving


def test_create_session_from_server_info_file(tmp_path: Path) -> None:
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
    ip = "127.0.0.1"
    port = get_free_port()
    server.add_insecure_port(f"{ip}:{port}")
    health_pb2_grpc.add_HealthServicer_to_server(MockHealthServicer(), server)
    scheme_eval_pb2_grpc.add_SchemeEvalServicer_to_server(
        MockSchemeEvalServicer(), server
    )
    server.start()
    server_info_file = tmp_path / "server_info.txt"
    server_info_file.write_text(f"{ip}:{port}\n12345")
    session = BaseSession.create_from_server_info_file(
        server_info_filepath=str(server_info_file), cleanup_on_exit=False
    )
    assert session.health_check_service.is_serving
    server.stop(None)
    session.exit()
    assert not session.health_check_service.is_serving


def test_create_session_from_server_info_file_with_wrong_password(
    tmp_path: Path,
) -> None:
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
    ip = "127.0.0.1"
    port = get_free_port()
    server.add_insecure_port(f"{ip}:{port}")
    scheme_eval_pb2_grpc.add_SchemeEvalServicer_to_server(
        MockSchemeEvalServicer(), server
    )
    server.start()
    server_info_file = tmp_path / "server_info.txt"
    server_info_file.write_text(f"{ip}:{port}\n1234")
    with pytest.raises(RuntimeError):
        session = BaseSession.create_from_server_info_file(
            server_info_filepath=str(server_info_file), cleanup_on_exit=False
        )
        session.scheme_eval.scheme_eval("")
        server.stop(None)
        session.exit()


def test_create_session_from_launch_fluent_by_passing_ip_and_port_and_password() -> (
    None
):
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
    ip = "127.0.0.1"
    port = get_free_port()
    server.add_insecure_port(f"{ip}:{port}")
    health_pb2_grpc.add_HealthServicer_to_server(MockHealthServicer(), server)
    scheme_eval_pb2_grpc.add_SchemeEvalServicer_to_server(
        MockSchemeEvalServicer(), server
    )
    server.start()
    session = launch_fluent(
        start_instance=False,
        ip=ip,
        port=port,
        cleanup_on_exit=False,
        mode="solver",
        password="12345",
    )
    # check a few dir elements
    session_dir = dir(session)
    for attr in ("field_data", "field_info", "setup", "solution"):
        assert attr in session_dir
    assert session.health_check_service.is_serving
    server.stop(None)
    session.exit()
    assert not session.health_check_service.is_serving


def test_create_session_from_launch_fluent_by_setting_ip_and_port_env_var(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
    ip = "127.0.0.1"
    port = get_free_port()
    server.add_insecure_port(f"{ip}:{port}")
    health_pb2_grpc.add_HealthServicer_to_server(MockHealthServicer(), server)
    scheme_eval_pb2_grpc.add_SchemeEvalServicer_to_server(
        MockSchemeEvalServicer(), server
    )
    server.start()
    monkeypatch.setenv("PYFLUENT_FLUENT_IP", ip)
    monkeypatch.setenv("PYFLUENT_FLUENT_PORT", str(port))
    session = launch_fluent(
        start_instance=False, cleanup_on_exit=False, mode="solver", password="12345"
    )
    # check a few dir elements
    session_dir = dir(session)
    for attr in ("field_data", "field_info"):
        assert attr in session_dir
    assert session.health_check_service.is_serving
    server.stop(None)
    session.exit()
    assert not session.health_check_service.is_serving


@pytest.mark.dev
@pytest.mark.fluent_232
def test_journal_creation(new_mesh_session):
    fd, file_path = tempfile.mkstemp(
        suffix=f"-{os.getpid()}.jou",
        prefix="pyfluent-",
        dir=str(pyfluent.EXAMPLES_PATH),
    )
    os.close(fd)

    prev_stat = Path(file_path).stat()
    prev_mtime = prev_stat.st_mtime
    prev_size = prev_stat.st_size

    session = new_mesh_session
    session.journal.start(file_path)
    session = session.switch_to_solver()
    session.journal.stop()

    new_stat = Path(file_path).stat()
    assert new_stat.st_mtime > prev_mtime
    assert new_stat.st_size > prev_size


@pytest.mark.skip("Failing in GitHub CI")
def test_old_style_session(with_launching_container):
    session = pyfluent.launch_fluent()
    case_path = download_file("mixing_elbow.cas.h5", "pyfluent/mixing_elbow")
    session.solver.root.file.read(file_type="case", file_name=case_path)
    session.solver.tui.report.system.sys_stats()
    session.exit()


def test_get_fluent_mode(new_mesh_session):
    session = new_mesh_session
    assert session.fluent_connection.get_current_fluent_mode() == "meshing"
    session = session.switch_to_solver()
    assert session.fluent_connection.get_current_fluent_mode() == "solver"


@pytest.mark.dev
@pytest.mark.fluent_232
@pytest.mark.skip("Failing in github")
def test_start_transcript_file_write(new_mesh_session):
    fd, file_path = tempfile.mkstemp(
        suffix=f"-{os.getpid()}.trn",
        prefix="pyfluent-",
        dir=str(pyfluent.EXAMPLES_PATH),
    )
    os.close(fd)

    prev_stat = Path(file_path).stat()
    prev_mtime = prev_stat.st_mtime
    prev_size = prev_stat.st_size

    session = new_mesh_session
    session.transcript.start(file_path)
    session = session.switch_to_solver()
    session.transcript.stop()

    new_stat = Path(file_path).stat()
    assert new_stat.st_mtime > prev_mtime
    assert new_stat.st_size > prev_size


@pytest.mark.fluent_231
def test_solverworkflow_in_solver_session(new_solver_session):
    solver = new_solver_session
    solver_dir = dir(solver)
    for attr in ("preferences", "solverworkflow", "tui", "workflow"):
        assert attr in solver_dir


@pytest.mark.dev
@pytest.mark.fluent_232
@pytest.mark.skip("Failing in github")
def test_read_case_using_lightweight_mode(with_launching_container):
    import_filename = examples.download_file(
        "mixing_elbow.cas.h5", "pyfluent/mixing_elbow"
    )
    solver = pyfluent.launch_fluent(
        case_filepath=import_filename, lightweight_mode=True
    )
    solver.setup.models.energy.enabled = False
    old_fluent_connection_id = id(solver.fluent_connection)
    while id(solver.fluent_connection) == old_fluent_connection_id:
        time.sleep(1)
    time.sleep(5)
    assert solver.setup.models.energy.enabled() == False
    solver.exit()



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from .view_name import view_name
class delete_view(Command):
    """
    Remove a view from the list.
    
    Parameters
    ----------
        view_name : str
            'view_name' child.
    
    """

    fluent_name = "delete-view"

    argument_names = \
        ['view_name']

    view_name: view_name = view_name
    """
    view_name argument of delete_view.
    """



import time

import pytest
from util.solver_workflow import new_solver_session  # noqa: F401

from ansys.fluent.core.fluent_connection import FluentConnection
from ansys.fluent.core.session_solver import Solver


def transcript(data):
    transcript.data = data


def run_transcript(i, ip, port, password):
    solver_session = Solver(
        FluentConnection(ip=ip, port=port, password=password, cleanup_on_exit=False)
    )
    solver_session.transcript.register_callback(transcript)

    transcript_counter = [0, 0]

    if i % 5 == 0:
        solver_session.scheme_eval.scheme_eval("(pp 'test)")
        check_transcript = True
        time.sleep(1)
        transcript_counter[0] += 1
    else:
        check_transcript = False

    if solver_session:
        solver_session.exit()
        if check_transcript:
            if not transcript.data:
                assert transcript.data == "test"
            else:
                assert transcript.data == "test"
                transcript_counter[1] += 1
        transcript("")

    return transcript_counter


@pytest.mark.dev
@pytest.mark.fluent_232
def test_transcript(new_solver_session):
    solver = new_solver_session
    ip = solver._channel_str.split(":")[0]
    port = int(solver._channel_str.split(":")[1])
    password = solver._metadata[0][-1]

    total_checked_transcript = 0
    passed_transcript = 0

    for i in range(100):
        transcript_counter = run_transcript(i, ip, port, password)
        total_checked_transcript += transcript_counter[0]
        passed_transcript += transcript_counter[1]

    if solver.get_fluent_version() >= "23.2.0":
        assert total_checked_transcript == passed_transcript
    else:
        assert total_checked_transcript > passed_transcript



"""
Module for creating quantities in terms of real values with units.
"""

from collections import OrderedDict


class _UnitsTable(object):
    """Units mappings"""

    fundamental_units = {
        "kg": "M",
        "g": "M",
        "lb": "M",
        "lbm": "M",
        "slug": "M",
        "m": "L",
        "cm": "L",
        "ft": "L",
        "in": "L",
        "s": "T",
        "A": "I",
        "K": "Temp",
        "C": "Temp",
        "F": "Temp",
        "R": "Temp",
        "mol": "N",
        "slugmol": "N",
        "cd": "Iv",
        "sr": "SAngle",
        "radian": "Angle",
        "degree": "Angle",
    }

    derived_units = {
        "N": "kg m s^-2",
        "Pa": "N m^-2",
        "W": "N m s^-1",
        "J": "N m",
        "V": "A ohm",
        "farad": "N m V^-2",
        "H": "N m A^-2",
        "S": "ohm^-1",
        "Wb": "N m A^-1",
        "T": "Wb m^-2",
        "dyne": "g cm s^-2",
        "erg": "dyne cm",
        "pdl": "lbm ft s^-2",
        "BTU": "J",
        "psi": "lbf in^-2",
        "lbf": "slug ft s^-2",
        "psf": "lbf ft^-2",
        "ohm": "kg m^2 s^-3 A^-2",
        "Hz": "s^-1",
        "cal": "J",
        "delta_K": "K",
        "delta_C": "C",
        "delta_F": "F",
        "delta_R": "R",
        "coulomb": "A s",
    }

    derived_units_with_conversion_factor = {
        "l": (0.001, "m^3"),
        "gal": (0.0037854117839999993, "m^3"),
        "BTU": (1055.056, "J"),
        "cal": (4.184, "J"),
    }

    multipliers = {
        "d": 10**-1,
        "c": 10**-2,
        "m": 10**-3,
        "u": 10**-6,
        "n": 10**-9,
        "p": 10**-12,
        "f": 10**-15,
        "a": 10**-18,
        "z": 10**-21,
        "y": 10**-24,
        "da": 10**1,
        "h": 10**2,
        "k": 10**3,
        "M": 10**6,
        "G": 10**9,
        "T": 10**12,
        "P": 10**15,
        "E": 10**18,
        "Z": 10**21,
        "Y": 10**24,
    }

    dimension_order = OrderedDict(
        [
            ("Mass", "M"),
            ("Length", "L"),
            ("Time", "T"),
            ("Temperature", "Temp"),
            ("Angle", "Angle"),
            ("ChemicalAmount", "N"),
            ("Light", "Iv"),
            ("Current", "I"),
            ("SolidAngle", "SAngle"),
        ]
    )

    conversion_map = {
        "kg": 1,
        "g": 0.001,
        "lb": 0.45359237,
        "lbm": 0.45359237,
        "slug": 14.59390293720637,
        "m": 1,
        "cm": 0.01,
        "ft": 0.30479999999999996,
        "in": 0.0254,
        "s": 1,
        "A": 1,
        "mol": 1,
        "slugmol": 14593.9,
        "cd": 1,
        "sr": 1,
        "radian": 1,
        "degree": 0.017453292519943295,
        "K": 1,
        "C": 1,
        "F": 0.5555555555555556,
        "R": 0.5555555555555556,
        "delta_K": 1,
        "delta_C": 1,
        "delta_F": 0.5555555555555556,
        "delta_R": 0.5555555555555556,
    }

    si_map = {
        "kg": "kg",
        "g": "kg",
        "lb": "kg",
        "lbm": "kg",
        "slug": "kg",
        "m": "m",
        "cm": "m",
        "ft": "m",
        "in": "m",
        "s": "s",
        "A": "A",
        "mol": "mol",
        "slugmol": "mol",
        "cd": "cd",
        "sr": "sr",
        "radian": "radian",
        "degree": "radian",
        "K": "K",
        "C": "K",
        "F": "K",
        "R": "K",
    }

    temperature_conversions = {
        # from : { to : formula }
        "C": {
            "K": lambda c: c + 273.15,
            "R": lambda c: (c + 273.15) * 1.8,
            "F": lambda c: c * 9 / 5 + 32,
            "C": lambda c: c,
        },
        "F": {
            "K": lambda f: (f - 32) * 5 / 9 + 273.15,
            "R": lambda f: f + 459.67,
            "F": lambda f: f,
            "C": lambda f: (f - 32) * 5 / 9,
        },
        "R": {
            "K": lambda r: r / 1.8,
            "R": lambda r: r,
            "F": lambda r: r - 459.67,
            "C": lambda r: (r - 491.67) / 1.8,
        },
        "K": {
            "K": lambda k: k,
            "R": lambda k: k * 1.8,
            "F": lambda k: k * 1.8 - 459.67,
            "C": lambda k: k - 273.15,
        },
    }


def get_si_conversion_factor(unit_str: str):
    """
    SI conversion factor.

    Parameters
    ----------
    unit_str : str
        To get required conversion factor

    Returns
    -------
    conversion_map
        A conversion map.
    """
    return (
        _UnitsTable.conversion_map[unit_str]
        if unit_str in _UnitsTable.conversion_map.keys()
        else 1
    )


def filter_multiplier(unit_str: str, predicate=None):
    """
    Separate multiplier from unit terms.

    Parameters
    ----------
    unit_str : str
        Unit of quantity.
    predicate
        Performs multiplier filtering.

    Returns
    -------
        matched
            A multiplier
    """
    result = False
    matched = ""
    for item in _UnitsTable.multipliers.keys():
        result = predicate(item, unit_str) if predicate else item == unit_str
        if result:
            matched = item
            result = True
            temp_unit_str = unit_str[len(item) :]
            if (
                temp_unit_str in _UnitsTable.fundamental_units
                or temp_unit_str in _UnitsTable.derived_units
            ):
                break
            else:
                result = False
                continue
    return result, matched


def remove_multiplier(unit_str: str) -> str:
    """
    Remove multiplier.

    Parameters
    ----------
    unit_str : str
        Unit of quantity.

    Returns
    -------
    str
        Unit with removed multiplier.
    """
    has_multiplier, prefix = filter_multiplier(
        unit_str, lambda item, unit: len(unit) > 0.0 and unit.startswith(item)
    )
    if has_multiplier:
        unit_str = unit_str[len(prefix) :]
    return unit_str


def is_temperature_quantity(dim_obj: dict) -> bool:
    """
    Checks if it is a temperature.

    Parameters
    ----------
    dim_obj : dict
        Mapping of fundamental quantities

    Returns
    -------
    bool
        Whether is it a temperature or not.
    """
    temp_dim = dim_obj["Temp"]
    total_dims_sum = sum([abs(val) for val in dim_obj.values()])
    return temp_dim == 1 and total_dims_sum == temp_dim


class Unit(object):
    """Unit postprocessing."""

    def __init__(self, unit_str):
        """
        Unit properties.

        Parameters
        ----------
        unit_str: str
            Unit of quantity
        """
        self._unit = unit_str
        self._si_multiplier = 1
        self._si_offset = 0
        self._si_unit = ""
        self._offset_power = 1
        self._compute_multipliers(unit_str, 1)
        self._compute_offset(unit_str)
        self._reduce_to_si_unit(self._si_unit)

    @property
    def user_unit(self):
        """User unit."""
        return self._unit

    @property
    def si_factor(self):
        """SI factor."""
        return self._si_multiplier

    @property
    def si_unit(self):
        """SI unit."""
        return self._si_unit

    @property
    def si_offset(self):
        """SI offset."""
        return self._si_offset

    @property
    def offset_power(self):
        """Offset power."""
        return self._offset_power

    def __call__(self):
        return self.user_unit

    def _reduce_to_si_unit(self, unit_str: str):
        """
        Reduce input unit to SI unit.

        Parameters
        ----------
        unit_str : str
            Unit of quantity
        """
        term_power_dict = OrderedDict()
        unit_split = unit_str.split(" ")

        for term in unit_split:
            if "^" in term:
                term_split = term.split("^")
                if term_split[0] in term_power_dict.keys():
                    term_power_dict[term_split[0]] += float(term_split[1])
                else:
                    term_power_dict[term_split[0]] = float(term_split[1])
            else:
                if term in term_power_dict.keys():
                    term_power_dict[term] += 1.0
                else:
                    term_power_dict[term] = 1.0

        self._si_unit = ""

        for key, power in term_power_dict.items():
            spacechar = " " if len(self._si_unit) > 0 else ""

            if power == 1.0:
                self._si_unit += spacechar + key
            elif power != 0.0:
                self._si_unit += (
                    spacechar
                    + key
                    + "^"
                    + str(int(power) if power.is_integer() else power)
                )
            else:
                self._si_unit += spacechar + key

    def _compute_multipliers(self, unit_str: str, power: float):
        """
        Compute multipliers.

        Parameters
        ----------
        unit_str: str
            Unit of quantity
        power: float
            Power of unit terms
        """
        if len(unit_str) == 0:
            return

        unit_list = unit_str.split(" ")

        for term in unit_list:
            unit_str = term
            term_power = 1

            if "^" in unit_str:
                unit_str, term_power = unit_str[: unit_str.index("^")], float(
                    unit_str[unit_str.index("^") + 1 :]
                )

            term_power *= power
            has_multiplier = not (
                unit_str in _UnitsTable.fundamental_units
                or unit_str in _UnitsTable.derived_units
            )

            if has_multiplier:
                _, prefix = filter_multiplier(
                    unit_str,
                    lambda item, unit: len(unit) > 0.0 and unit.startswith(item),
                )

                if len(prefix):
                    self._si_multiplier *= _UnitsTable.multipliers[prefix] ** term_power

                unit_str = remove_multiplier(unit_str)

            if unit_str in _UnitsTable.fundamental_units:
                spacechar = " " if len(self._si_unit) > 0 else ""

                if term_power == 1.0:
                    self._si_unit += spacechar + _UnitsTable.si_map[unit_str]
                elif term_power != 0.0:
                    self._si_unit += (
                        spacechar + _UnitsTable.si_map[unit_str] + "^" + str(term_power)
                    )
                elif term_power == 0.0:
                    self._si_unit += ""
                else:
                    self._si_unit += spacechar + _UnitsTable.si_map[unit_str]

                self._si_multiplier *= get_si_conversion_factor(unit_str) ** term_power

            elif unit_str in _UnitsTable.derived_units_with_conversion_factor:
                (
                    conversion_factor,
                    unit_str,
                ) = _UnitsTable.derived_units_with_conversion_factor[unit_str]
                self._si_multiplier *= conversion_factor**term_power
                self._compute_multipliers(unit_str, term_power)
            elif unit_str in _UnitsTable.derived_units:
                self._compute_multipliers(
                    _UnitsTable.derived_units[unit_str], term_power
                )

    @staticmethod
    def _power_sum(base, unit_str):
        """
        Sum of powers.

        Parameters
        ----------
        base: str
            Unit term
        unit_str: str
            Unit of quantity

        Returns
        -------
        power_sum: float
            Sum of powers with common unit terms
        """
        if unit_str in ["K", "C", "F", "R"]:
            if "^" in unit_str:
                return sum(
                    [
                        float(term.split("^")[1])
                        for term in unit_str.split(" ")
                        if term.split("^")[0] == base
                    ]
                )
            else:
                return 1.0

    def _compute_offset(self, unit_str: str) -> float:
        """
        Compute offset.

        Parameters
        ----------
        unit_str: str
            unit of quantity

        Returns
        -------
        offset: float
            Temperature conversion offset
        """
        if unit_str == "C":
            self._si_offset = 273.15
            self._offset_power = self._power_sum("C", unit_str)
        elif unit_str == "F":
            self._si_offset = 255.3722
            self._offset_power = self._power_sum("F", unit_str)
        else:
            self._si_offset = 0
            self._offset_power = 1.0

    def _quantity_type(self):
        """
        Quantity type base on temperature unit.

        Returns
        -------
        qty_type: str
            Type of temperature quantity
        """
        temperature_units = ["K", "C", "F", "R"]
        qty_type = None
        for temp_unit in temperature_units:
            unit_power = Unit._power_sum(temp_unit, self._unit)
            if temp_unit in self._unit and unit_power == 1.0:
                qty_type = "Temperature"
            elif "delta_" in self._unit:
                qty_type = "Temperature Difference"
            elif temp_unit in self._unit and unit_power not in [0.0, 1.0]:
                qty_type = "Temperature Difference"
        return qty_type


class Dimension(object):
    """Dimension of unit."""

    def __init__(self, unit_str):
        """
        Dimension properties.

        Parameters
        ----------
        unit_str: str
            Unit of quantity
        """
        self._dimensions = {
            "M": 0.0,
            "L": 0.0,
            "T": 0.0,
            "Temp": 0.0,
            "I": 0.0,
            "N": 0.0,
            "Iv": 0.0,
            "Angle": 0.0,
            "SAngle": 0.0,
        }
        self._parser(unit_str)

    def _add_or_update_dimension(self, dim_dict):
        """Add or update dimension."""
        for key in dim_dict.keys():
            self._dimensions[key] += dim_dict[key]

    def _parser(self, unit: str, power=1):
        """
        Unit string parser.

        Parameters
        ----------
        unit: str
            Unit term
        power: float
            Power of unit term
        """
        if len(unit) == 0:
            return

        unit_list = unit.split(" ")
        for term in unit_list:
            unit_dim = ""
            has_multiplier = not (
                term in _UnitsTable.fundamental_units
                or term in _UnitsTable.derived_units
            )
            unit_str = remove_multiplier(term) if has_multiplier else term
            term_power = 1

            if "^" in term:
                unit_str, term_power = term[: term.index("^")], float(
                    term[term.index("^") + 1 :]
                )
                has_multiplier = not (
                    unit_str in _UnitsTable.fundamental_units
                    or unit_str in _UnitsTable.derived_units
                )
                unit_str = remove_multiplier(unit_str) if has_multiplier else unit_str

            term_power *= power
            unit_dim = self._get_dim(unit_str, term_power)

            if unit_dim != None:
                self._add_or_update_dimension(unit_dim)

    def as_dict(self):
        """Dimension map."""
        return self._dimensions

    def _get_dim(self, unit_str, power):
        """Get dimensions"""
        if unit_str in _UnitsTable.fundamental_units:
            return {_UnitsTable.fundamental_units[unit_str]: power}
        elif unit_str in _UnitsTable.derived_units:
            self._parser(_UnitsTable.derived_units[unit_str], power)
        elif unit_str in _UnitsTable.derived_units_with_conversion_factor:
            _, unit_str = _UnitsTable.derived_units_with_conversion_factor[unit_str]
            self._parser(unit_str, power)
        else:
            raise ValueError("Not implemented")


def get_si_unit_from_dim(dim_list: list) -> str:
    """
    Get SI unit from dimensions.

    Parameters
    ----------
    dim_list: list
        list of dimensions

    Returns
    -------
    si_unit: str
        SI unit
    """
    dims_to_float = [float(dim) for dim in dim_list]
    dim_list = dims_to_float + ((9 - len(dims_to_float)) * [0.0])
    si_unit = ""
    dim_to_unit_map = UnitSystem("SI").base_units()

    for key, power in zip(_UnitsTable.dimension_order.values(), dim_list):
        unit_str = dim_to_unit_map[key]
        spacechar = " " if len(si_unit) > 0 else ""
        if power == 1.0:
            si_unit += spacechar + unit_str
        elif power != 0.0:
            si_unit += (
                spacechar
                + unit_str
                + "^"
                + str(int(power) if power.is_integer() else power)
            )
    return si_unit


class UnitSystem:
    """Unit systems."""

    _dim_to_unit_sys_map = {
        "SI": {
            "M": "kg",
            "L": "m",
            "T": "s",
            "Temp": "K",
            "I": "A",
            "N": "mol",
            "Iv": "cd",
            "Angle": "radian",
            "SAngle": "sr",
        },
        "CGS": {
            "M": "g",
            "L": "cm",
            "T": "s",
            "Temp": "K",
            "I": "A",
            "N": "mol",
            "Iv": "cd",
            "Angle": "radian",
            "SAngle": "sr",
        },
        "BT": {
            "M": "slug",
            "L": "ft",
            "T": "s",
            "Temp": "R",
            "I": "A",
            "N": "slugmol",
            "Iv": "cd",
            "Angle": "radian",
            "SAngle": "sr",
        },
    }

    _supported_unit_sys = ("SI", "CGS", "BT")

    def __init__(self, unit_sys):
        """
        Basic unit systems.

        Parameters
        ----------
        unit_sys: str
            System name
        """
        self._unit_system = unit_sys.upper()

        if self._unit_system not in UnitSystem._supported_unit_sys:
            raise ValueError(
                "Unsupported unit system, only 'SI', 'CGS', 'BT' is allowed."
            )

    def _get_unit_from_dim(self, dim_list: list) -> str:
        """
        Get unit from dimensions.

        Parameters
        ----------
        dim_list: list
            List of dimensions

        Returns
        -------
        unit: str
            Unit using dimensions
        """
        unit = ""

        base_units = UnitSystem._dim_to_unit_sys_map[self._unit_system]
        for key, power in zip(_UnitsTable.dimension_order.values(), dim_list):
            spacechar = " " if len(unit) > 0 else ""
            if power == 1:
                unit += spacechar + base_units[key]
            elif power != 0:
                unit += (
                    spacechar
                    + base_units[key]
                    + "^"
                    + str(int(power) if power.is_integer() else power)
                )
        return unit

    def convert(self, quantity):
        """
        Convert to a system.

        Parameters
        ----------
        quantity: Quantity
            Quantity instance

        Returns
        -------
        quantity: Quantity
            Quantity instance
        """
        if not isinstance(quantity, Quantity):
            raise TypeError("Not instance of Quantity.")
        dims = quantity.get_dimensions_list()
        unit_str = self._get_unit_from_dim(dims)
        return quantity.to(unit_str)

    def base_units(self):
        """Base units."""
        return UnitSystem._dim_to_unit_sys_map[self._unit_system]


class QuantityError(ValueError):
    """Custom quantity errors."""

    def __init__(self, from_unit, to_unit):
        """
        Quantity errors with custom messages.

        Parameters
        ----------
        from_unit: str
            Unit of quantity
        to_unit: str
            Desired conversion unit
        """
        self.from_unit = from_unit
        self.to_unit = to_unit

    def __str__(self):
        return f"{self.unit} and {self.to_unit} have incompatible dimensions."


unit_quantity_map = {
    "Mass": "kg",
    "Length": "m",
    "Time": "s",
    "Temperature": "K",
    "Current": "A",
    "SubstanceAmount": "mol",
    "Light": "cd",
    "Angle": "radian",
    "SolidAngle": "sr",
    "Acceleration": "m s^-2",
    "Angular Acceleration": "radian s^-2",
    "Angular Velocity": "radian s^-1",
    "Area": "m^2",
    "Capacitance": "farad",
    "Compressibility": "Pa^-1",
    "Concentration": "m^-3",
    "Contact Resistance": "m^2 s kg^-1",
    "Current Transfer Coefficient": "A m^-2 V^-1",
    "Decay Constant": "s^-1",
    "Density": "kg m^-3",
    "Density Derivative": "m^-2 s^2",
    "Density Derivative wrt Pressure": "m^-2 s^2",
    "Density Derivative wrt Temperature": "kg m^-3 K^-1",
    "Dielectric Contact Resistance": "farad^-1 m^2",
    "Dynamic Viscosity": "Pa s",
    "Electric Charge": "A s",
    "Electric Charge Density": "A s m^-3",
    "Electric Charge Transfer Coefficient": "farad m^-2",
    "Electric Conductance Per Unit Area": "S m^-2",
    "Electric Current Density": "A m^-2",
    "Electric Current Source": "A m^-3",
    "Electric Field": "V m^-1",
    "Electric Flux Density": "coulomb m^-2",
    "Electrical Conductance": "S",
    "Electrical Conductivity": "S m^-1",
    "Electrical Contact Resistance": "S^-1 m^2",
    "Electrical Permittivity": "A s V^-1 m^-1",
    "Electrical Resistance": "ohm",
    "Electrical Resistivity": "ohm m",
    "Energy Density by Mass": "J kg^-1",
    "Energy Source": "W m^-3",
    "Energy Source Coefficient": "W m^-3 K^-1",
    "Enthalpy Variance": "m^4 s^-4",
    "Epsilon": "m^2 s^-3",
    "Epsilon Flux": "W m^-2 s^-1",
    "Epsilon Flux Coefficient": "kg m^-2 s^-2",
    "Epsilon Source": "W m^-3 s^-1",
    "Epsilon Source Coefficient": "kg m^-3 s^-2",
    "Flame Surface Density Source": "m^-1 s^-1",
    "Force": "N",
    "Force Density": "N m^-3",
    "Force Intensity": "N m^-1",
    "Force Per Angular Unit": "N radian^-1",
    "Fracture Energy": "J m^-2",
    "Fracture Energy Rate": "J m^-2 s^-1",
    "Frequency": "Hz",
    "Gasket Stiffness": "Pa m^-1",
    "Heat Flux": "W m^-2",
    "Heat Flux in": "W m^-2",
    "Heat Generation": "W m^-3",
    "Heat Rate": "W",
    "Heat Transfer Coefficient": "W m^-2 K^-1",
    "Impulse": "N s",
    "Impulse Per Angular Unit": "N s radian^-1",
    "Inductance": "H",
    "Interphase Transfer Coefficient": "kg m^-2 s^-1",
    "InvTemp1": "K^-1",
    "InvTemp2": "K^-2",
    "InvTemp3": "K^-3",
    "InvTemp4": "K^-4",
    "Inverse Angle": "radian^-1",
    "Inverse Area": "m^-2",
    "Inverse Length": "m^-1",
    "Inverse Stress": "Pa^-1",
    "Kinematic Diffusivity": "m^2 s^-1",
    "MAPDL Enthalpy": "J m^-3",
    "Magnetic Field": "A m^-1",
    "Magnetic Field Intensity": "A m^-1",
    "Magnetic Flux": "Wb",
    "Magnetic Flux Density": "T",
    "Magnetic Induction": "T",
    "Magnetic Permeability": "H m^-1",
    "Magnetic Potential": "T m",
    "Mass Concentration": "kg m^-3",
    "Mass Concentration Rate": "kg m^-3 s^-1",
    "Mass Flow": "kg s^-1",
    "Mass Flow Rate Per Area": "kg s^-1 m^-2",
    "Mass Flow Rate Per Length": "kg s^-1 m^-1",
    "Mass Flow Rate Per Volume": "kg s^-1 m^-3",
    "Mass Flow in": "kg s^-1",
    "Mass Flux": "kg s^-1 m^-2",
    "Mass Flux Coefficient": "kg s^-1 m^-2",
    "Mass Flux Pressure Coefficient": "kg s^-1 m^-2 Pa^-1",
    "Mass Fraction": "kg kg^-1",
    "Mass Per Area": "kg m^-2",
    "Mass Source": "kg s^-1 m^-3",
    "Mass Source Coefficient": "kg s^-1 m^-3 Pa^-1",
    "Material Impedance": "kg m^-2 s^-1",
    "Molar Concentration": "mol m^-3",
    "Molar Concentration Henry Coefficient": "Pa m^3 mol^-1",
    "Molar Concentration Rate": "mol m^-3 s^-1",
    "Molar Energy": "J mol^-1",
    "Molar Entropy": "J mol^-1 K^-1",
    "Molar Fraction": "mol mol^-1",
    "Molar Mass": "kg kmol^-1",
    "Molar Volume": "m^3 mol^-1",
    "Moment": "N m",
    "Moment of Inertia of Area": "m^2 m^2",
    "Moment of Inertia of Mass": "kg m^2",
    "Momentum Source": "kg m^-2 s^-2",
    "Momentum Source Lin Coeff": "kg m^-3 s^-1",
    "Momentum Source Quad Coeff": "kg m^-4",
    "Normalized Value": "m m^-1",
    "Number Source": "m^-3 s^-1",
    "Omega Source": "kg m^-3 s^-2",
    "PSD Acceleration": "m^2 s^-4 Hz^-1",
    "PSD Displacement": "m^2 Hz^-1",
    "PSD Force": "N^2 Hz^-1",
    "PSD Moment": "N^2 m^2 Hz^-1",
    "PSD Pressure": "Pa^2 Hz^-1",
    "PSD Strain": "m^2 m^-2 Hz^-1",
    "PSD Stress": "Pa^2 Hz^-1",
    "PSD Velocity": "m^2 s^-2 Hz^-1",
    "Per Mass": "kg^-1",
    "Per Mass Flow": "s kg^-1",
    "Per Time": "s^-1",
    "Per Time Cubed": "s^-3",
    "Per Time Squared": "s^-2",
    "Power Spectral Density": "W Hz^-1",
    "Pressure": "Pa",
    "Pressure Derivative wrt Temperature": "Pa K^-1",
    "Pressure Derivative wrt Volume": "Pa kg m^-3",
    "Pressure Gradient": "Pa m^-1",
    "Relative Permeability": "H H^-1",
    "Relative Permittivity": "farad farad^-1",
    "Rotational Damping": "N m s radian^-1",
    "Rotational Stiffness": "N m radian^-1",
    "Section Modulus": "m^3",
    "Seebeck Coefficient": "V K^-1",
    "Shear Strain": "radian",
    "Shear Strain Rate": "s^-1",
    "Shock Velocity": "s m^-1",
    "Soot Cross Coefficient": "m^3 mol^-1 s^-1",
    "Soot PX Facto": "mol kg^-1 s^-1",
    "Specific": "kg kg^-1",
    "Specific Concentration": "mol kg^-1",
    "Specific Energy": "J kg^-1",
    "Specific Enthalpy": "J kg^-1",
    "Specific Entropy": "J kg^-1 K^-1",
    "Specific Flame Surface Density": "m^2 kg^-1",
    "Specific Heat Capacity": "J kg^-1 K^-1",
    "Specific Volume": "kg^-1 m^3",
    "Specific Weight": "N m^-3",
    "Stiffness": "N m^-1",
    "Strain": "m m^-1",
    "Strength": "Pa",
    "Stress": "Pa",
    "Stress Intensity Factor": "Pa m^0.5",
    "Stress Per Temperature": "Pa K^-1",
    "Surface Charge Density": "A s m^-2",
    "Surface Force Density": "N m^-2",
    "Surface Power Density": "W m^-2",
    "Surface Tension": "N m^-1",
    "Temperature Difference": "delta_K",
    "Temperature Gradient": "K m^-1",
    "Temperature Variance": "K^2",
    "Temperature Variance Source": "kg m^-3 s^-1 K^2",
    "Thermal Capacitance": "J K^-1",
    "Thermal Conductance": "W K^-1",
    "Thermal Conductivity": "W m^-1 K^-1",
    "Thermal Contact Resistance": "W^-1 m^2 K",
    "Thermal Expansivity": "K^-1",
    "Torque": "N m",
    "Torsional Spring Constant": "N m radian^-1",
    "Total Mass Source Pressure Coefficient": "kg s^-1 Pa^-1",
    "Total Radiative Intensity": "W m^-2 sr^-1",
    "Translational Damping": "N s m^-1",
    "Turbulent Heat Flux": "m^3 s^-3",
    "Velocity": "m s^-1",
    "Volume": "m^3",
    "Volumetric": "kg m^-3",
    "Volumetric Flow": "m^3 s^-1",
    "Volumetric Flow in": "m^3 s^-1",
    "Warping Factor": "m^6",
}


def get_unit_from_map(quantity_map_from_settings_api):
    """
    Get unit from quantity map.

    Parameters
    ----------
    quantity_map_from_settings_api: dict
        Map of quantity and it's unit

    Returns
    -------
    unit: str
        Unit of quantity using quantity map
    """
    terms = []
    term_power_list = []
    final_unit = ""
    for unit in quantity_map_from_settings_api.keys():
        unit_str = unit_quantity_map[unit]
        unit_term_list = unit_str.split(" ") if " " in unit_str else unit_str
        term_power_list.append(unit_term_list)
    power_list = list(
        map(lambda x: float(x), list(quantity_map_from_settings_api.values()))
    )
    for item in range(len(power_list)):
        if isinstance(term_power_list[item], str) and power_list[item] == 1.0:
            terms.append(term_power_list[item])
        elif isinstance(term_power_list[item], list) and power_list[item] == 1.0:
            for term in term_power_list[item]:
                terms.append(term)
        elif isinstance(term_power_list[item], str) and power_list[item] != 0.0:
            terms.append(
                term_power_list[item]
                + "^"
                + str(
                    int(power_list[item])
                    if power_list[item].is_integer()
                    else power_list[item]
                )
            )
        elif isinstance(term_power_list[item], list) and power_list[item] != 0.0:
            for term in term_power_list[item]:
                if "^" in term:
                    term_split = term.split("^")
                    resulting_power = float(term_split[1]) * power_list[item]
                    terms.append(
                        term_split[0]
                        + "^"
                        + str(
                            int(resulting_power)
                            if resulting_power.is_integer()
                            else resulting_power
                        )
                    )
                else:
                    terms.append(
                        term
                        + "^"
                        + str(
                            int(power_list[item])
                            if power_list[item].is_integer()
                            else power_list[item]
                        )
                    )
    for term in terms:
        final_unit += (" " if len(final_unit) > 0 else "") + term

    unit_term_list = final_unit.split(" ")
    unit_power_list = [
        terms.split("^") if "^" in terms else terms for terms in unit_term_list
    ]

    term_power_map = {}
    for term in unit_power_list:
        if isinstance(term, str):
            if term in term_power_map.keys():
                term_power_map[term] += 1.0
            else:
                term_power_map[term] = 1.0
        if isinstance(term, list):
            if term[0] in term_power_map.keys():
                term_power_map[term[0]] += float(term[1])
            else:
                term_power_map[term[0]] = float(term[1])

    unit_terms = []
    for term, power in term_power_map.items():
        if power == 1.0:
            unit_terms.append(term)
        elif power != 0.0:
            unit_terms.append(
                term + "^" + str(int(power) if power.is_integer() else power)
            )

    unit = ""
    for term in unit_terms:
        unit += (" " if len(unit) > 0 else "") + term
    return unit


class Quantity(float):
    """This class instantiates physical quantities using their real values and
    units. All the instances of this class are converted to base SI units system to have
    consistency in all arithmetic operations.

    Attributes
    ----------
    value: Real value
        Value of quantity is stored as float.

    unit: Unit string
        Unit of quantity is stored as string.

    Methods
    -------
    get_dimensions_list()
        Extracts dimensions from unit.

    is_dimensionless()
        Determines type of quantity.

    to(to_unit)
        Converts to given unit string.

    Returns
    -------
    Quantity instance.
    """

    def __init__(self, real_value, unit_str=None, quantity_map=None, dimensions=None):
        """
        Quantity using unit string, quantity map or dimensions.

        Parameters
        ----------
        real_value: float
            Value of quantity
        unit_str: str
            Unit of quantity
        quantity_map: dict
            Map of quantity and it's unit
        dimensions: list
            array of dimensions

        """
        self._value = float(real_value)
        if (
            (unit_str and quantity_map)
            or (unit_str and dimensions)
            or (quantity_map and dimensions)
        ):
            raise ValueError("unit or quantity_map or dimensions is allowed.")
        if quantity_map:
            unit_str = get_unit_from_map(quantity_map)
        if dimensions:
            if len(dimensions) > 9:
                raise ValueError("Number of dimensions should be <= 9")
            unit_str = get_si_unit_from_dim(dim_list=dimensions)
        self._unit = Unit(unit_str)
        self._dimension = Dimension(unit_str)
        self._si_value = (
            self._unit.si_factor * self._value + self._unit.si_offset
        ) ** self._unit.offset_power
        self._si_unit = self._unit.si_unit
        self._type = self._unit._quantity_type()

    def __new__(cls, real_value, unit_str=None, quantity_map=None, dimensions=None):
        if unit_str and quantity_map:
            raise ValueError("Either unit or quantity_map is allowed.")
        if quantity_map:
            unit_str = get_unit_from_map(quantity_map)
        if dimensions:
            if len(dimensions) > 9:
                raise ValueError("Number of dimensions should be <= 9")
            unit_str = get_si_unit_from_dim(dim_list=dimensions)
        _unit = Unit(unit_str)
        return float.__new__(
            cls, (_unit.si_factor * real_value + _unit.si_offset) ** _unit.offset_power
        )

    @property
    def value(self):
        """Value of quantity."""
        return self._value

    @property
    def unit(self):
        """Unit of quantity."""
        return self._unit.user_unit

    @value.setter
    def value(self, val):
        self._value = val

    @property
    def dimension(self):
        """Dimensions of quantity."""
        return self._dimension

    @property
    def type(self):
        """Quantity type."""
        return self._type

    @type.setter
    def type(self, type_str):
        self._type = type_str

    def is_dimensionless(self):
        """Determines whether quantity is dimensionsless or not."""
        return all([value == 0 for value in self.get_dimensions_list()])

    def get_dimensions_list(self):
        """Dimensions list."""
        dims = self._dimension.as_dict()
        return [
            dims[_UnitsTable.dimension_order[key]]
            for key in _UnitsTable.dimension_order.keys()
        ]

    def to(self, to_unit_str):
        """
        Convert quantity to desired unit.

        Parameters
        ----------
        to_unit_str: str
            Desired unit of conversion

        Returns
        -------
        temp_quantity: Quantity
            Instance of converted quantity
        """
        if not isinstance(to_unit_str, str):
            raise TypeError("Unit should be of 'str' type.")

        temp_quantity = Quantity(1, to_unit_str)

        curr_unit_dim_obj = self.dimension
        to_unit_dim_obj = temp_quantity.dimension

        if curr_unit_dim_obj.as_dict() != to_unit_dim_obj.as_dict():
            raise QuantityError(self.unit, to_unit_str)
        else:
            pass

        if is_temperature_quantity(curr_unit_dim_obj.as_dict()):
            return self._get_converted_temperature(to_unit_str)
        else:
            curr_unit_obj = self._unit
            to_unit_obj = temp_quantity._unit
            temp_quantity.value = (
                curr_unit_obj.si_factor / to_unit_obj.si_factor
            ) * self.value
            return temp_quantity

    def _get_converted_temperature(self, to_unit_str):
        """Get converted temperature."""
        from_unit_str = self.unit

        if "^" in from_unit_str:
            from_unit_str = from_unit_str.split("^")[0]

        if "^" in to_unit_str:
            to_unit_str = to_unit_str.split("^")[0]

        _, prefix = filter_multiplier(
            from_unit_str, lambda item, unit: len(unit) > 0.0 and unit.startswith(item)
        )
        from_unit_multiplier = (
            _UnitsTable.multipliers[prefix] if len(prefix) != 0 else 1.0
        )

        _, prefix = filter_multiplier(
            to_unit_str, lambda item, unit: len(unit) > 0.0 and unit.startswith(item)
        )
        to_unit_multiplier = (
            _UnitsTable.multipliers[prefix] if len(prefix) != 0 else 1.0
        )

        from_key = remove_multiplier(from_unit_str)
        to_key = remove_multiplier(to_unit_str)

        return Quantity(
            _UnitsTable.temperature_conversions[from_key][to_key](
                from_unit_multiplier * self.value
            )
            / to_unit_multiplier,
            to_unit_str,
        )

    def _get_si_unit(self, other, func):
        """Get SI unit."""
        curr_dim = self.get_dimensions_list()
        other_dim = other.get_dimensions_list()
        temp_dim = [func(curr_dim[i], other_dim[i]) for i in range(len(other_dim))]
        temp_unit = get_si_unit_from_dim(temp_dim)
        return temp_unit

    def __pow__(self, exponent):
        new_dims = list(map(lambda x: x * exponent, self.get_dimensions_list()))
        new_si_unit = get_si_unit_from_dim(new_dims)
        new_si_value = pow(self._si_value, exponent)
        return Quantity(new_si_value, new_si_unit)

    def __str__(self):
        return f'({self.value}, "{self.unit}")'

    def __repr__(self):
        return f'Quantity ({self.value}, "{self.unit}")'

    def __mul__(self, other):
        if isinstance(other, Quantity):
            temp_value = self._si_value * other._si_value
            temp_unit = self._get_si_unit(other, lambda x, y: x + y)
            return Quantity(temp_value, temp_unit)
        elif isinstance(other, int) or isinstance(other, float):
            if self.type == "Temperature Difference":
                result = Quantity(float(self) * float(other), "delta_K")
                result.type = "Temperature Difference"
                return result
            return Quantity(self._si_value * other, self._si_unit)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, Quantity):
            temp_value = self._si_value / other._si_value
            temp_unit = self._get_si_unit(other, lambda x, y: x - y)
            return Quantity(temp_value, temp_unit)
        elif isinstance(other, int) or isinstance(other, float):
            temp = Quantity(self._si_value / other, self._si_unit)
            return temp

    def __rtruediv__(self, other):
        if not isinstance(other, Quantity) and isinstance(other, float):
            return Quantity(other / self._si_value, self._si_unit)
        else:
            return other / self

    def validate_matching_dimensions(self, other):
        """Validate dimensions of quantities."""
        if isinstance(other, Quantity) and (
            self.get_dimensions_list() != other.get_dimensions_list()
        ):
            raise ValueError("Incompatible dimensions.")
        elif (
            (not self.is_dimensionless())
            and (not isinstance(other, Quantity))
            and isinstance(other, (float, int))
        ):
            raise TypeError("Incompatible quantities.")

    def __add__(self, other):
        self.validate_matching_dimensions(other)
        temp_types = ["Temperature", "Temperature Difference"]
        if self.type in temp_types and other.type in temp_types:
            result = Quantity(float(self) + float(other), "K")
            result.type = "Temperature"
            return result
        temp_value = float(self) + float(other)
        return Quantity(temp_value, self._si_unit)

    def __radd__(self, other):
        return Quantity(other, "") + self

    def __sub__(self, other):
        self.validate_matching_dimensions(other)
        temp_types = ["Temperature", "Temperature Difference"]
        if self.type in temp_types and other.type in temp_types:
            result = Quantity(float(self) - float(other), "delta_K")
            result.type = "Temperature Difference"
            return result
        temp_value = float(self) - float(other)
        return Quantity(temp_value, self._si_unit)

    def __rsub__(self, other):
        return Quantity(other, "") - self

    def __neg__(self):
        return Quantity(-self.value, self.unit)

    def __gt__(self, other):
        self.validate_matching_dimensions(other)
        return float(self) > float(other)

    def __ge__(self, other):
        self.validate_matching_dimensions(other)
        return float(self) >= float(other)

    def __lt__(self, other):
        self.validate_matching_dimensions(other)
        return float(self) < float(other)

    def __le__(self, other):
        self.validate_matching_dimensions(other)
        return float(self) <= float(other)

    def __eq__(self, other):
        self.validate_matching_dimensions(other)
        return float(self) == float(other)

    def __neq__(self, other):
        self.validate_matching_dimensions(other)
        return float(self) != float(other)



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class ocw_ship_vel_spec(String):
    """
    'ocw_ship_vel_spec' child.
    """

    fluent_name = "ocw-ship-vel-spec"




#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from .project_filename import project_filename
from .load_case import load_case
class open(Command):
    """
    Open project.
    
    Parameters
    ----------
        project_filename : str
            'project_filename' child.
        load_case : bool
            'load_case' child.
    
    """

    fluent_name = "open"

    argument_names = \
        ['project_filename', 'load_case']

    project_filename: project_filename = project_filename
    """
    project_filename argument of open.
    """
    load_case: load_case = load_case
    """
    load_case argument of open.
    """



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from .project_filename import project_filename
from .convert_to_managed import convert_to_managed
class save_as_copy(Command):
    """
    Save As Project.
    
    Parameters
    ----------
        project_filename : str
            'project_filename' child.
        convert_to_managed : bool
            'convert_to_managed' child.
    
    """

    fluent_name = "save-as-copy"

    argument_names = \
        ['project_filename', 'convert_to_managed']

    project_filename: project_filename = project_filename
    """
    project_filename argument of save_as_copy.
    """
    convert_to_managed: convert_to_managed = convert_to_managed
    """
    convert_to_managed argument of save_as_copy.
    """



import pytest
from util.fixture_fluent import load_static_mixer_case  # noqa: F401

from ansys.fluent.core.services.reduction import _locn_names_and_objs

load_static_mixer_case_2 = load_static_mixer_case


def _test_locn_extraction(solver1, solver2):
    locns = _locn_names_and_objs(["inlet1"])
    assert locns == [["inlet1", ["inlet1"]]]

    all_bcs = solver1.setup.boundary_conditions
    locns = _locn_names_and_objs(all_bcs)
    assert locns == [
        ["interior--fluid", all_bcs],
        ["outlet", all_bcs],
        ["inlet1", all_bcs],
        ["inlet2", all_bcs],
        ["wall", all_bcs],
    ]

    locns = _locn_names_and_objs([all_bcs["inlet1"]])
    assert locns == [["inlet1", all_bcs["inlet1"]]]

    all_bcs = solver1.setup.boundary_conditions
    all_bcs2 = solver2.setup.boundary_conditions
    locns = _locn_names_and_objs([all_bcs, all_bcs2])
    assert locns == [
        ["interior--fluid", all_bcs],
        ["outlet", all_bcs],
        ["inlet1", all_bcs],
        ["inlet2", all_bcs],
        ["wall", all_bcs],
        ["interior--fluid", all_bcs2],
        ["outlet", all_bcs2],
        ["inlet1", all_bcs2],
        ["inlet2", all_bcs2],
        ["wall", all_bcs2],
    ]


def _test_context(solver):
    solver.solution.initialization.hybrid_initialize()

    assert solver.reduction.area(
        locations=[solver.setup.boundary_conditions.velocity_inlet["inlet1"]],
        ctxt=solver,
    )

    assert solver.reduction.area(locations=["inlet1"], ctxt=solver)


def _test_area_average(solver):
    solver.solution.initialization.hybrid_initialize()
    solver.setup.named_expressions["test_expr_1"] = {}
    solver.setup.named_expressions[
        "test_expr_1"
    ].definition = "AreaAve(AbsolutePressure, ['inlet1'])"
    expr_val = solver.setup.named_expressions["test_expr_1"].get_value()
    assert type(expr_val) == float and expr_val != 0.0
    val = solver.reduction.area_average(
        expression="AbsolutePressure",
        locations=solver.setup.boundary_conditions.velocity_inlet,
    )
    assert val == expr_val
    solver.setup.named_expressions.pop(key="test_expr_1")


def _test_min(solver1, solver2):
    vmag = solver1.setup.boundary_conditions["inlet1"].vmag.value()
    solver1.setup.boundary_conditions["inlet1"].vmag = 0.9 * vmag
    solver2.setup.boundary_conditions["inlet1"].vmag = 1.1 * vmag
    solver1.solution.initialization.hybrid_initialize()
    solver2.solution.initialization.hybrid_initialize()
    solver1.setup.named_expressions["test_expr_1"] = {}
    test_expr1 = solver1.setup.named_expressions["test_expr_1"]
    test_expr1.definition = "sqrt(VelocityMagnitude)"
    solver1.setup.named_expressions["test_expr_2"] = {}
    test_expr2 = solver1.setup.named_expressions["test_expr_2"]
    test_expr2.definition = "minimum(test_expr_2, ['outlet'])"
    expected_result = test_expr2.get_value()
    result = solver1.reduction.minimum(
        expression=test_expr1.definition(),
        locations=[
            solver1.setup.boundary_conditions["outlet"],
            solver2.setup.boundary_conditions["outlet"],
        ],
    )
    # assert result == expected_result
    solver1.setup.named_expressions.pop(key="test_expr_1")
    solver1.setup.named_expressions.pop(key="test_expr_2")


def _test_count(solver):
    solver.solution.initialization.hybrid_initialize()
    solver.setup.named_expressions["test_expr_1"] = {}
    solver.setup.named_expressions["test_expr_1"].definition = "Count(['inlet1'])"
    expr_val_1 = solver.setup.named_expressions["test_expr_1"].get_value()
    solver.setup.named_expressions["test_expr_1"].definition = "Count(['inlet2'])"
    expr_val_2 = solver.setup.named_expressions["test_expr_1"].get_value()
    solver.setup.named_expressions[
        "test_expr_1"
    ].definition = "Count(['inlet1', 'inlet2'])"
    expr_val_3 = solver.setup.named_expressions["test_expr_1"].get_value()
    assert expr_val_3 == expr_val_1 + expr_val_2
    red_val_1 = solver.reduction.count(
        locations=[solver.setup.boundary_conditions.velocity_inlet["inlet1"]]
    )
    red_val_2 = solver.reduction.count(
        locations=[solver.setup.boundary_conditions.velocity_inlet["inlet2"]]
    )
    red_val_3 = solver.reduction.count(
        locations=[solver.setup.boundary_conditions.velocity_inlet]
    )
    assert red_val_1 == expr_val_1
    assert red_val_2 == expr_val_2
    assert red_val_3 == expr_val_3
    solver.setup.named_expressions.pop(key="test_expr_1")


def _test_centroid(solver):
    solver.solution.initialization.hybrid_initialize()
    solver.setup.named_expressions["test_expr_1"] = {}
    solver.setup.named_expressions["test_expr_1"].definition = "Centroid(['inlet1'])"
    expr_val_1 = solver.setup.named_expressions["test_expr_1"].get_value()
    solver.setup.named_expressions["test_expr_1"].definition = "Centroid(['inlet2'])"
    expr_val_2 = solver.setup.named_expressions["test_expr_1"].get_value()
    solver.setup.named_expressions[
        "test_expr_1"
    ].definition = "Centroid(['inlet1', 'inlet2'])"
    expr_val_3 = solver.setup.named_expressions["test_expr_1"].get_value()
    red_val_1 = solver.reduction.centroid(
        locations=[solver.setup.boundary_conditions.velocity_inlet["inlet1"]]
    )
    red_val_2 = solver.reduction.centroid(
        locations=[solver.setup.boundary_conditions.velocity_inlet["inlet2"]]
    )
    red_val_3 = solver.reduction.centroid(
        locations=[solver.setup.boundary_conditions.velocity_inlet]
    )
    assert [red_val_1.x, red_val_1.y, red_val_1.z] == expr_val_1
    assert [red_val_2.x, red_val_2.y, red_val_2.z] == expr_val_2
    assert [red_val_3.x, red_val_3.y, red_val_3.z] == expr_val_3
    solver.setup.named_expressions.pop(key="test_expr_1")


def _test_area_integrated_average(solver1, solver2):
    solver1.solution.initialization.hybrid_initialize()
    solver2.solution.initialization.hybrid_initialize()

    solver1.setup.named_expressions["test_expr_1"] = {}
    solver1.setup.named_expressions[
        "test_expr_1"
    ].definition = "AreaInt(AbsolutePressure, ['inlet1'])"
    expr_val_1 = solver1.setup.named_expressions["test_expr_1"].get_value()
    solver1.setup.named_expressions[
        "test_expr_1"
    ].definition = "AreaInt(AbsolutePressure, ['inlet2'])"
    expr_val_2 = solver1.setup.named_expressions["test_expr_1"].get_value()
    solver1.setup.named_expressions[
        "test_expr_1"
    ].definition = "AreaInt(AbsolutePressure, ['inlet1', 'inlet2'])"
    expr_val_3 = solver1.setup.named_expressions["test_expr_1"].get_value()
    assert expr_val_3 - (expr_val_1 + expr_val_2) <= 0.000000001

    red_val_1 = solver1.reduction.area_integral(
        expression="AbsolutePressure",
        locations=[solver1.setup.boundary_conditions.velocity_inlet["inlet1"]],
    )
    red_val_2 = solver1.reduction.area_integral(
        expression="AbsolutePressure",
        locations=[solver1.setup.boundary_conditions.velocity_inlet["inlet2"]],
    )
    red_val_3 = solver1.reduction.area_integral(
        expression="AbsolutePressure",
        locations=[solver1.setup.boundary_conditions.velocity_inlet],
    )

    assert red_val_1 == expr_val_1
    assert red_val_2 == expr_val_2
    assert red_val_3 == expr_val_3

    solver2.setup.named_expressions["test_expr_1"] = {}
    solver2.setup.named_expressions[
        "test_expr_1"
    ].definition = "AreaInt(AbsolutePressure, ['inlet1'])"
    expr_val_4 = solver2.setup.named_expressions["test_expr_1"].get_value()
    solver2.setup.named_expressions[
        "test_expr_1"
    ].definition = "AreaInt(AbsolutePressure, ['inlet2'])"
    expr_val_5 = solver2.setup.named_expressions["test_expr_1"].get_value()
    solver2.setup.named_expressions[
        "test_expr_1"
    ].definition = "AreaInt(AbsolutePressure, ['inlet1', 'inlet2'])"
    expr_val_6 = solver2.setup.named_expressions["test_expr_1"].get_value()
    assert expr_val_6 - (expr_val_4 + expr_val_5) <= 0.000000001

    red_val_4 = solver2.reduction.area_integral(
        expression="AbsolutePressure",
        locations=[solver2.setup.boundary_conditions.velocity_inlet["inlet1"]],
    )
    red_val_5 = solver2.reduction.area_integral(
        expression="AbsolutePressure",
        locations=[solver2.setup.boundary_conditions.velocity_inlet["inlet2"]],
    )
    red_val_6 = solver2.reduction.area_integral(
        expression="AbsolutePressure",
        locations=[solver2.setup.boundary_conditions.velocity_inlet],
    )

    assert red_val_4 == expr_val_4
    assert red_val_5 == expr_val_5
    assert red_val_6 == expr_val_6

    red_val_7 = solver2.reduction.area_integral(
        expression="AbsolutePressure",
        locations=[
            solver1.setup.boundary_conditions.velocity_inlet,
            solver2.setup.boundary_conditions.velocity_inlet,
        ],
    )

    assert red_val_7 - (expr_val_3 + expr_val_6) <= 0.000000001

    solver1.setup.named_expressions.pop(key="test_expr_1")


def _test_error_handling(solver):
    with pytest.raises(RuntimeError) as msg:
        solver.reduction.area_average(
            expression="AbsoluteVelocity",  # This is a wrong expression intentionally passed
            locations=solver.setup.boundary_conditions.velocity_inlet,
        )

    assert (
        msg.value.args[0]
        == "The last request could not be completed because there is error in server."
    )


def _test_force(solver):
    solver.solution.initialization.hybrid_initialize()
    solver.setup.named_expressions["test_expr_1"] = {}
    solver.setup.named_expressions["test_expr_1"].definition = "Force(['wall'])"
    expr_val_1 = solver.setup.named_expressions["test_expr_1"].get_value()

    red_total_force = solver.reduction.force(
        locations=[solver.setup.boundary_conditions.wall]
    )
    red_pressure_force = solver.reduction.pressure_force(locations=["wall"])
    red_viscous_force = solver.reduction.viscous_force(
        locations=[solver.setup.boundary_conditions.wall]
    )

    assert [red_total_force.x, red_total_force.y, red_total_force.z] == expr_val_1

    assert red_pressure_force.x + red_viscous_force.x == red_total_force.x

    assert red_pressure_force.y + red_viscous_force.y == red_total_force.y

    assert red_pressure_force.z + red_viscous_force.z == red_total_force.z

    solver.setup.named_expressions.pop(key="test_expr_1")


def _test_moment(solver):
    solver.solution.initialization.hybrid_initialize()
    solver.setup.named_expressions["test_expr_1"] = {}
    solver.setup.named_expressions[
        "test_expr_1"
    ].definition = "Moment(Force(['wall']),['wall'])"
    expr_val_1 = solver.setup.named_expressions["test_expr_1"].get_value()

    solver.setup.named_expressions[
        "test_expr_1"
    ].definition = "Moment(['inlet1'],['wall'])"
    expr_val_2 = solver.setup.named_expressions["test_expr_1"].get_value()

    red_moment_force = solver.reduction.moment(
        expression="Force(['wall'])", locations=[solver.setup.boundary_conditions.wall]
    )

    red_moment_location = solver.reduction.moment(
        expression="['inlet1']",
        locations=[solver.setup.boundary_conditions.wall],
    )

    assert [red_moment_force.x, red_moment_force.y, red_moment_force.z] == expr_val_1
    assert [
        red_moment_location.x,
        red_moment_location.y,
        red_moment_location.z,
    ] == expr_val_2

    solver.setup.named_expressions.pop(key="test_expr_1")


@pytest.mark.dev
@pytest.mark.fluent_232
def test_reductions(load_static_mixer_case, load_static_mixer_case_2) -> None:
    solver1 = load_static_mixer_case
    solver2 = load_static_mixer_case_2
    _test_context(solver1)
    _test_locn_extraction(solver1, solver2)
    _test_area_average(solver1)
    _test_min(solver1, solver2)
    _test_count(solver1)
    _test_centroid(solver1)
    _test_area_integrated_average(solver1, solver2)
    _test_error_handling(solver1)
    _test_force(solver1)
    _test_moment(solver1)



from ansys.fluent.core.datamodel_231.PMFileManagement import (
    Root as pmfilemanagement_root,
)
from ansys.fluent.core.datamodel_231.PartManagement import Root as partmanagement_root
from ansys.fluent.core.datamodel_231.meshing import Root as meshing_root
from ansys.fluent.core.datamodel_231.preferences import Root as preferences_root
from ansys.fluent.core.datamodel_231.workflow import Root as workflow_root
from ansys.fluent.core.solver.tui_231 import main_menu

class Meshing:
    @property
    def tui(self) -> main_menu: ...
    @property
    def meshing(self) -> meshing_root: ...
    @property
    def workflow(self) -> workflow_root: ...
    @property
    def PartManagement(self) -> partmanagement_root: ...
    @property
    def PMFileManagement(self) -> pmfilemanagement_root: ...
    @property
    def preferences(self) -> preferences_root: ...



import pytest

import ansys.fluent.core as pyfluent


@pytest.mark.dev
@pytest.mark.fluent_231
def test_icing_session(with_launching_container):
    icing_session = pyfluent.launch_fluent(mode=pyfluent.LaunchMode.SOLVER_ICING)
    assert "icing" in dir(icing_session)



"""Module to manage datamodel cache."""


from collections import defaultdict
from typing import Any, Dict, List, Union

from ansys.api.fluent.v0.variant_pb2 import Variant

StateType = Union[
    bool,
    int,
    float,
    str,
    List[bool],
    List[int],
    List[float],
    List[str],
    List["StateType"],
    Dict[str, "StateType"],
]


class DataModelCache:
    """Class to manage datamodel cache."""

    class Empty:
        """Class representing unassigned cached state."""

    @staticmethod
    def is_unassigned(state: Any) -> bool:
        """Check whether a cached state is unassigned.

        Parameters
        ----------
        state : Any
            state

        Returns
        -------
        bool
            whether a cached state is unassigned
        """
        return state is DataModelCache.Empty

    rules_str_to_cache = defaultdict(dict)
    rules_str_to_config = {}

    @staticmethod
    def get_config(rules: str, name: str) -> Any:
        """Get datamodel cache configuration value.

        Parameters
        ----------
        rules : str
            datamodel rules
        name : str
            configuration name

        Returns
        -------
        Any
            configuration value
        """
        return DataModelCache.rules_str_to_config.get(rules, {}).get(name, False)

    @staticmethod
    def set_config(rules: str, name: str, value: Any):
        """Set datamodel cache configuration value.

        Parameters
        ----------
        rules : str
            datamodel rules
        name : str
            configuration name
        value : Any
            configuration value
        """
        if rules not in DataModelCache.rules_str_to_config:
            DataModelCache.rules_str_to_config[rules] = {}
        DataModelCache.rules_str_to_config[rules][name] = value

    @staticmethod
    def _update_cache_from_variant_state(
        rules: str, source: Dict[str, StateType], key: str, state: Variant, updaterFn
    ):
        if state.HasField("bool_state"):
            updaterFn(source, key, state.bool_state)
        elif state.HasField("int64_state"):
            updaterFn(source, key, state.int64_state)
        elif state.HasField("double_state"):
            updaterFn(source, key, state.double_state)
        elif state.HasField("string_state"):
            updaterFn(source, key, state.string_state)
        elif state.HasField("bool_vector_state"):
            updaterFn(source, key, state.bool_vector_state.item)
        elif state.HasField("int64_vector_state"):
            updaterFn(source, key, state.int64_vector_state.item)
        elif state.HasField("double_vector_state"):
            updaterFn(source, key, state.double_vector_state.item)
        elif state.HasField("string_vector_state"):
            updaterFn(source, key, state.string_vector_state.item)
        elif state.HasField("variant_vector_state"):
            updaterFn(source, key, [])
            for item in state.variant_vector_state.item:
                DataModelCache._update_cache_from_variant_state(
                    rules, source, key, item, lambda d, k, v: d[k].append(v)
                )
        elif state.HasField("variant_map_state"):
            internal_names_as_keys = DataModelCache.get_config(
                rules, "internal_names_as_keys"
            )
            if ":" in key:
                type_, iname = key.split(":", maxsplit=1)
                for k1, v1 in source.items():
                    if (internal_names_as_keys and k1 == key) or (
                        (not internal_names_as_keys)
                        and isinstance(v1, dict)
                        and v1.get("__iname__") == iname
                    ):
                        key = k1
                        break
                else:  # new named object
                    if internal_names_as_keys:
                        source[key] = {}
                    else:
                        name = state.variant_map_state.item["_name_"].string_state
                        key = f"{type_}:{name}"
                        source[key] = {"__iname__": iname}
            else:
                if key not in source:
                    source[key] = {}
            source = source[key]
            for k, v in state.variant_map_state.item.items():
                DataModelCache._update_cache_from_variant_state(
                    rules, source, k, v, dict.__setitem__
                )

    @staticmethod
    def update_cache(rules: str, state: Variant, deleted_paths: List[str]):
        """Update datamodel cache from streamed state.

        Parameters
        ----------
        rules : str
            datamodel rules
        state : Variant
            streamed state
        deleted_paths : List[str]
            list of deleted paths
        """
        cache = DataModelCache.rules_str_to_cache[rules]
        internal_names_as_keys = DataModelCache.get_config(
            rules, "internal_names_as_keys"
        )
        for deleted_path in deleted_paths:
            comps = [x for x in deleted_path.split("/") if x]
            sub_cache = cache
            for i, comp in enumerate(comps):
                if ":" in comp:
                    _, iname = comp.split(":", maxsplit=1)
                    key_to_del = None
                    for k, v in sub_cache.items():
                        if (internal_names_as_keys and k == comp) or (
                            (not internal_names_as_keys)
                            and isinstance(v, dict)
                            and v.get("__iname__") == iname
                        ):
                            if i == len(comps) - 1:
                                key_to_del = k
                            else:
                                sub_cache = v
                            break
                    else:
                        break
                    if key_to_del:
                        del sub_cache[key_to_del]
                else:
                    if comp in sub_cache:
                        sub_cache = sub_cache[comp]
                    else:
                        break
        for k, v in state.variant_map_state.item.items():
            DataModelCache._update_cache_from_variant_state(
                rules, cache, k, v, dict.__setitem__
            )

    @staticmethod
    def _dm_path_comp(comp):
        return ":".join(comp if comp[1] else comp[0])

    @staticmethod
    def _dm_path_comp_list(obj):
        return [DataModelCache._dm_path_comp(comp) for comp in obj.path]

    @staticmethod
    def get_state(rules: str, obj: object) -> Any:
        """Retrieve state from datamodel cache

        Parameters
        ----------
        rules : str
            datamodel rules
        obj : object
            datamodel object

        Returns
        -------
        _type_
            _description_
        """
        cache = DataModelCache.rules_str_to_cache[rules]
        if not len(cache):
            return DataModelCache.Empty
        path_components = DataModelCache._dm_path_comp_list(obj)
        for path_component in path_components:
            cache = cache.get(path_component, None)
            if not cache:
                return DataModelCache.Empty
        return cache

    @staticmethod
    def _set_state_at_path(cache, path, value):
        if len(path) == 0:
            return
        path_component = path[0]
        if len(path) == 1:
            cache[path_component] = value
        else:
            next_cache = cache.get(path_component, None)
            if not next_cache:
                next_cache = cache[path_component] = dict()
            DataModelCache._set_state_at_path(next_cache, path[1:], value)

    @staticmethod
    def set_state(rules: str, obj: object, value: Any):
        """Set datamodel cache state

        Parameters
        ----------
        rules : str
            datamodel rules
        obj : object
            datamodel object
        value : Any
            state
        """
        DataModelCache._set_state_at_path(
            DataModelCache.rules_str_to_cache[rules],
            DataModelCache._dm_path_comp_list(obj),
            value,
        )



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class partially_catalytic_material(String):
    """
    'partially_catalytic_material' child.
    """

    fluent_name = "partially-catalytic-material"




"""Interceptor classes to use with gRPC services."""

import logging
from typing import Any

from google.protobuf.json_format import MessageToDict
import grpc

from ansys.fluent.core.services.batch_ops import BatchOps

network_logger = logging.getLogger("ansys.fluent.networking")


class TracingInterceptor(grpc.UnaryUnaryClientInterceptor):
    """Interceptor class to trace gRPC calls."""

    def __init__(self):
        """__init__ method of TracingInterceptor class."""
        super().__init__()

    def _intercept_call(
        self,
        continuation: Any,
        client_call_details: grpc.ClientCallDetails,
        request: Any,
    ):
        network_logger.debug(
            "GRPC_TRACE: rpc = %s, request = %s",
            client_call_details.method,
            MessageToDict(request),
        )
        response = continuation(client_call_details, request)
        if not response.exception():
            network_logger.debug(
                "GRPC_TRACE: response = %s",
                MessageToDict(response.result()),
            )
        return response

    def intercept_unary_unary(
        self,
        continuation: Any,
        client_call_details: grpc.ClientCallDetails,
        request: Any,
    ) -> Any:
        """Intercept unary-unary call for tracing."""
        return self._intercept_call(continuation, client_call_details, request)


class BatchedFuture(grpc.Future):
    """Class implementing gRPC.Future interface.

    An instance of BatchedFuture is returned if the gRPC method is
    queued to be executed in batch later.
    """

    def __init__(self, result_cls):
        """__init__ method of BatchedFuture class."""
        self._result_cls = result_cls

    def cancel(self):
        """Attempts to cancel the computation."""
        return False

    def cancelled(self):
        """Describes whether the computation was cancelled."""
        return False

    def running(self):
        """Describes whether the computation is taking place."""
        return False

    def done(self):
        """Describes whether the computation has taken place."""
        return True

    def result(self, timeout=None):
        """Returns the result of the computation or raises its exception."""
        return self._result_cls()

    def exception(self, timeout=None):
        """Return the exception raised by the computation."""
        return None

    def traceback(self, timeout=None):
        """Access the traceback of the exception raised by the computation."""
        return None

    def add_done_callback(self, fn):
        """Adds a function to be called at completion of the computation."""
        pass


class BatchInterceptor(grpc.UnaryUnaryClientInterceptor):
    """Interceptor class to batch gRPC calls."""

    def __init__(self):
        """__init__ method of BatchInterceptor class."""
        super().__init__()

    def _intercept_call(
        self,
        continuation: Any,
        client_call_details: grpc.ClientCallDetails,
        request: Any,
    ):
        batchOps = BatchOps.instance()
        if batchOps and batchOps.batching:
            qual_method = client_call_details.method
            package_and_service, method = qual_method.lstrip("/").split("/")
            package, service = package_and_service.rsplit(".", 1)
            op = batchOps.add_op(package, service, method, request)
            if op.queued:
                return BatchedFuture(op.response_cls)

        return continuation(client_call_details, request)

    def intercept_unary_unary(
        self,
        continuation: Any,
        client_call_details: grpc.ClientCallDetails,
        request: Any,
    ) -> Any:
        """Intercept unary-unary call for batch operation."""
        return self._intercept_call(continuation, client_call_details, request)



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class volume_fraction_courant_number(Real):
    """
    'volume_fraction_courant_number' child.
    """

    fluent_name = "volume-fraction-courant-number"




#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class delete_all_compute_nodes(Boolean):
    """
    'delete_all_compute_nodes' child.
    """

    fluent_name = "delete-all-compute-nodes?"




import threading
from typing import Callable

from ansys.api.fluent.v0 import datamodel_se_pb2
from ansys.fluent.core.services.datamodel_se import _convert_variant_to_value
from ansys.fluent.core.streaming_services.streaming import StreamingService


class DatamodelEvents(StreamingService):
    """Encapsulates a datamodel events streaming service."""

    def __init__(self, service):
        """Instantiate DatamodelEvents."""
        super().__init__(
            target=DatamodelEvents._process_streaming,
            streaming_service=service,
        )
        self._cbs = {}
        service.event_streaming = self
        self._lock = threading.RLock()

    def register_callback(self, tag: str, obj, cb: Callable):
        """Register a callback."""
        with self._lock:
            self._cbs[tag] = obj, cb

    def unregister_callback(self, tag: str):
        """Unregister a callback."""
        with self._lock:
            self._cbs.pop(tag, None)

    def _process_streaming(self, started_evt):
        """Processes datamodel events."""
        responses = self._streaming_service.begin_event_streaming(started_evt)
        while True:
            try:
                response: datamodel_se_pb2.EventResponse = next(responses)
                with self._lock:
                    self._streaming = True
                    for tag, cb in self._cbs.items():
                        if tag == response.tag:
                            if response.HasField("createdEventResponse"):
                                childtype = response.createdEventResponse.childtype
                                childname = response.createdEventResponse.childname
                                child = getattr(cb[0], childtype)[childname]
                                cb[1](child)
                            elif (
                                response.HasField("modifiedEventResponse")
                                or response.HasField("deletedEventResponse")
                                or response.HasField("affectedEventResponse")
                                or response.HasField("attributeChangedEventResponse")
                                or response.HasField(
                                    "commandAttributeChangedEventResponse"
                                )
                            ):
                                cb[1](cb[0])
                            elif response.HasField("commandExecutedEventResponse"):
                                command = response.commandExecutedEventResponse.command
                                args = _convert_variant_to_value(
                                    response.commandExecutedEventResponse.args
                                )
                                cb[1](cb[0], command, args)
            except StopIteration:
                break



from ansys.api.fluent.v0 import datamodel_se_pb2
from ansys.fluent.core.streaming_services.streaming import StreamingService


class DatamodelStream(StreamingService):
    """Encapsulates a datamodel streaming service."""

    def __init__(self, service):
        """Instantiate DatamodelStream."""
        super().__init__(
            target=DatamodelStream._process_streaming,
            streaming_service=service,
        )

    def _process_streaming(self, started_evt):
        """Processes datamodel events."""
        responses = self._streaming_service.begin_streaming(started_evt)
        while True:
            try:
                response: datamodel_se_pb2.DataModelResponse = next(responses)
                with self._lock:
                    self._streaming = True
                    for _, cb_list in self._service_callbacks.items():
                        state = response.state if hasattr(response, "state") else None
                        deleted_paths = getattr(response, "deletedpaths", None)
                        cb_list[0](state=state, deleted_paths=deleted_paths)
            except StopIteration:
                break



import os
import subprocess
import time

import psutil
from util.solver_workflow import (  # noqa: F401
    new_solver_session,
    new_solver_session_no_transcript,
)

import ansys.fluent.core as pyfluent
from ansys.fluent.core.examples import download_file


def _read_case(session):
    case_path = download_file("Static_Mixer_main.cas.h5", "pyfluent/static_mixer")
    session.file.read(file_type="case", file_name=case_path)


def test_session_starts_transcript_by_default(new_solver_session) -> None:
    session = new_solver_session

    def print_transcript(transcript):
        print_transcript.called = True
        if transcript:
            print_transcript.transcript = transcript

    print_transcript.called = False
    print_transcript.transcript = None

    session.transcript.register_callback(print_transcript)

    _read_case(session=session)

    assert print_transcript.called
    assert print_transcript.transcript


def test_session_starts_no_transcript_if_disabled(
    new_solver_session_no_transcript,
) -> None:
    session = new_solver_session_no_transcript

    def print_transcript(transcript):
        print_transcript.called = True

    print_transcript.called = False

    session.transcript.start(write_to_stdout=False)

    _read_case(session=session)

    assert not print_transcript.called


def get_container_ids_set():
    proc = subprocess.Popen(["docker", "ps", "-q"], stdout=subprocess.PIPE)
    output_bytes = proc.stdout.read()
    output_str = output_bytes.decode()
    ids = output_str.strip().split()
    return set(ids)


def test_server_exits_when_session_goes_out_of_scope(with_launching_container) -> None:
    def f(get_pid):
        session = pyfluent.launch_fluent(mode="solver")
        if get_pid:
            f.server_pid = session.scheme_eval.scheme_eval("(%cx-process-id)")

    if os.getenv("PYFLUENT_START_INSTANCE") == "0":
        containers_before = get_container_ids_set()
        f(get_pid=False)
        time.sleep(10)
        containers_after = get_container_ids_set()
        new_containers = containers_after - containers_before
        assert not new_containers
    else:
        f(get_pid=True)
        time.sleep(10)
        assert not psutil.pid_exists(f.server_pid)


def test_server_does_not_exit_when_session_goes_out_of_scope(
    with_launching_container,
) -> None:
    def f():
        session = pyfluent.launch_fluent(mode="solver", cleanup_on_exit=False)
        f.server_pid = session.scheme_eval.scheme_eval("(%cx-process-id)")

    if os.getenv("PYFLUENT_START_INSTANCE") == "0":
        containers_before = get_container_ids_set()
        f()
        time.sleep(10)
        containers_after = get_container_ids_set()
        new_containers = containers_after - containers_before
        assert new_containers
        for container in new_containers:
            subprocess.Popen(["docker", "stop", container])
    else:
        f()
        time.sleep(10)
        assert psutil.pid_exists(f.server_pid)
        psutil.Process(f.server_pid).kill()



from ansys.fluent.core.datamodel_231.PMFileManagement import (
    Root as pmfilemanagement_root,
)
from ansys.fluent.core.datamodel_231.PartManagement import Root as partmanagement_root
from ansys.fluent.core.datamodel_231.meshing import Root as meshing_root
from ansys.fluent.core.datamodel_231.preferences import Root as preferences_root
from ansys.fluent.core.datamodel_231.workflow import Root as workflow_root
from ansys.fluent.core.solver.tui_231 import main_menu

class PureMeshing:
    @property
    def tui(self) -> main_menu: ...
    @property
    def meshing(self) -> meshing_root: ...
    @property
    def workflow(self) -> workflow_root: ...
    @property
    def PartManagement(self) -> partmanagement_root: ...
    @property
    def PMFileManagement(self) -> pmfilemanagement_root: ...
    def transfer_mesh_to_solvers(
        self,
        solvers,
        file_type: str = ...,
        file_name_stem: str = ...,
        num_files_to_try: int = ...,
        clean_up_mesh_file: bool = ...,
        overwrite_previous: bool = ...,
    ): ...



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class min_step_change_factor(Real):
    """
    Set Minimum Step Change Factor.
    """

    fluent_name = "min-step-change-factor"




#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from .axis_direction_component_child import axis_direction_component_child

class band_diffuse_frac(NamedObject[axis_direction_component_child], _CreatableNamedObjectMixin[axis_direction_component_child]):
    """
    'band_diffuse_frac' child.
    """

    fluent_name = "band-diffuse-frac"

    child_object_type: axis_direction_component_child = axis_direction_component_child
    """
    child_object_type of band_diffuse_frac.
    """



"""A package providing Fluent's Solver and Meshing capabilities in Python."""

import logging.config
import os
import pydoc

import appdirs
import yaml

from ansys.fluent.core._version import __version__  # noqa: F401
from ansys.fluent.core.launcher.launcher import (  # noqa: F401
    FluentVersion,
    LaunchMode,
    launch_fluent,
)
from ansys.fluent.core.services.batch_ops import BatchOps  # noqa: F401
from ansys.fluent.core.session import BaseSession as Fluent  # noqa: F401
from ansys.fluent.core.utils import fldoc
from ansys.fluent.core.utils.setup_for_fluent import setup_for_fluent  # noqa: F401

_VERSION_INFO = "Build date: May 19, 2023 17:23 UTC ShaID: 162e819"
"""Global variable indicating the version of the PyFluent package - Empty by default"""

_THIS_DIRNAME = os.path.dirname(__file__)
_README_FILE = os.path.normpath(os.path.join(_THIS_DIRNAME, "docs", "README.rst"))

if os.path.exists(_README_FILE):
    with open(_README_FILE, encoding="utf8") as f:
        __doc__ = f.read()


def version_info() -> str:
    """Method returning the version of PyFluent being used.

    Returns
    -------
    str
        The PyFluent version being used.

    Notes
    -------
    Only available in packaged versions. Otherwise it will return __version__.
    """
    return _VERSION_INFO if _VERSION_INFO is not None else __version__


file_path = os.path.abspath(__file__)
file_dir = os.path.dirname(file_path)
yaml_path = os.path.join(file_dir, "logging_config.yaml")

# Load the logging configuration from a YAML file
with open(yaml_path, "rt") as f:
    config = yaml.safe_load(f)

# Configure the logging system
logging.config.dictConfig(config)

# Setup data directory
USER_DATA_PATH = appdirs.user_data_dir(appname="ansys_fluent_core", appauthor="Ansys")
EXAMPLES_PATH = os.path.join(USER_DATA_PATH, "examples")

BUILDING_GALLERY = False

# Set this to False to stop automatically inferring and setting REMOTING_SERVER_ADDRESS
INFER_REMOTING_IP = True

# Time in second to wait for response for each ip while inferring remoting ip
INFER_REMOTING_IP_TIMEOUT_PER_IP = 2

pydoc.text.docother = fldoc.docother.__get__(pydoc.text, pydoc.TextDoc)

# Whether to use datamodel state caching
DATAMODEL_USE_STATE_CACHE = True

# Whether to use datamodel attribute caching
DATAMODEL_USE_ATTR_CACHE = True

# Whether stream and cache commands state
DATAMODEL_USE_NOCOMMANDS_DIFF_STATE = True



from time import sleep

import pytest
from util.meshing_workflow import new_mesh_session  # noqa: F401

from ansys.api.fluent.v0 import datamodel_se_pb2
import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples
from ansys.fluent.core.services.datamodel_se import (
    _convert_variant_to_value,
    convert_path_to_se_path,
)
from ansys.fluent.core.services.streaming import StreamingService
from ansys.fluent.core.streaming_services.datamodel_streaming import DatamodelStream


@pytest.mark.dev
@pytest.mark.fluent_232
def test_event_subscription(new_mesh_session):
    session = new_mesh_session
    session.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    tags = [
        "/workflow/created/TaskObject",
        "/workflow/modified/TaskObject:TaskObject1/Arguments",
        "/workflow/deleted/TaskObject:TaskObject1",
        "/workflow/affected/TaskObject:TaskObject1",
        "/workflow/affected/TaskObject",
        "/workflow/attribute_changed/TaskObject:TaskObject1/TaskList/isActive",
        "/workflow/command_attribute_changed/InitializeWorkflow/arguments",
        "/workflow/command_executed/InitializeWorkflow",
    ]
    request = datamodel_se_pb2.SubscribeEventsRequest()
    e1 = request.eventrequest.add(rules="workflow")
    e1.createdEventRequest.parentpath = ""
    e1.createdEventRequest.childtype = "TaskObject"
    e2 = request.eventrequest.add(rules="workflow")
    e2.modifiedEventRequest.path = "TaskObject:TaskObject1/Arguments"
    e3 = request.eventrequest.add(rules="workflow")
    e3.deletedEventRequest.path = "TaskObject:TaskObject1"
    e4 = request.eventrequest.add(rules="workflow")
    e4.affectedEventRequest.path = "TaskObject:TaskObject1"
    e5 = request.eventrequest.add(rules="workflow")
    e5.affectedEventRequest.path = ""
    e5.affectedEventRequest.subtype = "TaskObject"
    e6 = request.eventrequest.add(rules="workflow")
    e6.attributeChangedEventRequest.path = "TaskObject:TaskObject1/TaskList"
    e6.attributeChangedEventRequest.attribute = "isActive"
    e7 = request.eventrequest.add(rules="workflow")
    e7.commandAttributeChangedEventRequest.path = ""
    e7.commandAttributeChangedEventRequest.command = "InitializeWorkflow"
    e7.commandAttributeChangedEventRequest.attribute = "arguments"
    e8 = request.eventrequest.add(rules="workflow")
    e8.commandExecutedEventRequest.path = ""
    e8.commandExecutedEventRequest.command = "InitializeWorkflow"
    response = session.fluent_connection.datamodel_service_se.subscribe_events(request)
    assert all(
        [
            r.status == datamodel_se_pb2.STATUS_SUBSCRIBED and r.tag == t
            for r, t in zip(response.response, tags)
        ]
    )

    request = datamodel_se_pb2.UnsubscribeEventsRequest()
    request.tag.extend(tags)
    response = session.fluent_connection.datamodel_service_se.unsubscribe_events(
        request
    )
    assert all(
        [
            r.status == datamodel_se_pb2.STATUS_UNSUBSCRIBED and r.tag == t
            for r, t in zip(response.response, tags)
        ]
    )


@pytest.mark.dev
@pytest.mark.fluent_232
def test_add_on_child_created(new_mesh_session):
    meshing = new_mesh_session
    child_paths = []
    subscription = meshing.workflow.add_on_child_created(
        "TaskObject", lambda obj: child_paths.append(convert_path_to_se_path(obj.path))
    )
    assert child_paths == []
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    sleep(5)
    assert len(child_paths) > 0
    child_paths.clear()
    subscription.unsubscribe()
    meshing.workflow.InitializeWorkflow(WorkflowType="Fault-tolerant Meshing")
    sleep(5)
    assert child_paths == []


@pytest.mark.dev
@pytest.mark.fluent_232
def test_add_on_deleted(new_mesh_session):
    meshing = new_mesh_session
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    data = []
    subscription = meshing.workflow.TaskObject["Import Geometry"].add_on_deleted(
        lambda obj: data.append(convert_path_to_se_path(obj.path))
    )
    assert data == []
    meshing.workflow.InitializeWorkflow(WorkflowType="Fault-tolerant Meshing")
    sleep(5)
    assert len(data) > 0


@pytest.mark.dev
@pytest.mark.fluent_232
def test_add_on_changed(new_mesh_session):
    meshing = new_mesh_session
    task_list = meshing.workflow.Workflow.TaskList
    assert isinstance(task_list(), list)
    assert len(task_list()) == 0
    data = []
    subscription = task_list.add_on_changed(lambda obj: data.append(len(obj())))
    assert data == []
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    sleep(5)
    assert len(data) > 0
    assert data[0] > 0
    data.clear()
    subscription.unsubscribe()
    meshing.workflow.InitializeWorkflow(WorkflowType="Fault-tolerant Meshing")
    sleep(5)
    assert data == []


@pytest.mark.dev
@pytest.mark.fluent_232
def test_add_on_affected(new_mesh_session):
    meshing = new_mesh_session
    data = []
    subscription = meshing.workflow.Workflow.add_on_affected(
        lambda obj: data.append(True)
    )
    assert data == []
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    sleep(5)
    assert len(data) > 0
    assert data[0] == True
    data.clear()
    subscription.unsubscribe()
    meshing.workflow.InitializeWorkflow(WorkflowType="Fault-tolerant Meshing")
    sleep(5)
    assert data == []


@pytest.mark.dev
@pytest.mark.fluent_232
def test_add_on_affected_at_type_path(new_mesh_session):
    meshing = new_mesh_session
    data = []
    subscription = meshing.workflow.add_on_affected_at_type_path(
        "TaskObject", lambda obj: data.append(True)
    )
    assert data == []
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    sleep(5)
    assert len(data) > 0
    assert data[0] == True
    data.clear()
    subscription.unsubscribe()
    meshing.workflow.InitializeWorkflow(WorkflowType="Fault-tolerant Meshing")
    sleep(5)
    assert data == []


@pytest.mark.dev
@pytest.mark.fluent_232
def test_add_on_command_executed(new_mesh_session):
    meshing = new_mesh_session
    data = []
    subscription = meshing.meshing.add_on_command_executed(
        "ImportGeometry", lambda obj, command, args: data.append(True)
    )
    assert data == []
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    import_filename = examples.download_file(
        "mixing_elbow.pmdb", "pyfluent/mixing_elbow"
    )
    meshing.meshing.ImportGeometry(FileName=import_filename)
    sleep(5)
    assert len(data) > 0
    assert data[0] == True
    data.clear()
    subscription.unsubscribe()
    meshing.meshing.ImportGeometry(FileName=import_filename)
    sleep(5)
    assert data == []


@pytest.fixture
def disable_datamodel_cache(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(pyfluent, "DATAMODEL_USE_STATE_CACHE", False)


@pytest.mark.dev
@pytest.mark.fluent_232
def test_datamodel_streaming_full_diff_state(disable_datamodel_cache, new_mesh_session):
    meshing = new_mesh_session
    datamodel_service_se = meshing.datamodel_service_se
    data_model_request = datamodel_se_pb2.DataModelRequest()
    data_model_request.rules = "meshing"
    datamodel_streaming = StreamingService(
        stub=datamodel_service_se._stub,
        request=data_model_request,
        metadata=datamodel_service_se._metadata,
    )
    stream = DatamodelStream(datamodel_streaming)
    stream.start()

    def cb(state, deleted_paths):
        if state:
            state = _convert_variant_to_value(state)
        cb.states.append(state)

    cb.states = []

    stream.register_callback(cb)

    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    import_filename = examples.download_file(
        "mixing_elbow.pmdb", "pyfluent/mixing_elbow"
    )
    meshing.meshing.ImportGeometry(FileName=import_filename)
    sleep(5)
    assert "ImportGeometry:ImportGeometry1" in (y for x in cb.states for y in x)


@pytest.mark.dev
@pytest.mark.fluent_232
def test_datamodel_streaming_no_commands_diff_state(
    disable_datamodel_cache, new_mesh_session
):
    meshing = new_mesh_session
    datamodel_service_se = meshing.datamodel_service_se
    data_model_request = datamodel_se_pb2.DataModelRequest()
    data_model_request.rules = "meshing"
    data_model_request.diffstate = datamodel_se_pb2.DIFFSTATE_NOCOMMANDS
    datamodel_streaming = StreamingService(
        stub=datamodel_service_se._stub,
        request=data_model_request,
        metadata=datamodel_service_se._metadata,
    )
    stream = DatamodelStream(datamodel_streaming)
    stream.start()

    def cb(state, deleted_paths):
        if state:
            state = _convert_variant_to_value(state)
        cb.states.append(state)

    cb.states = []

    stream.register_callback(cb)

    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    import_filename = examples.download_file(
        "mixing_elbow.pmdb", "pyfluent/mixing_elbow"
    )
    meshing.meshing.ImportGeometry(FileName=import_filename)
    sleep(5)
    assert "ImportGeometry:ImportGeometry1" not in (y for x in cb.states for y in x)


@pytest.mark.dev
@pytest.mark.fluent_232
def test_get_object_names_wtm(new_mesh_session):
    meshing = new_mesh_session

    assert not meshing.workflow.TaskObject.get_object_names()

    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")

    child_object_names = [
        "Import Geometry",
        "Add Local Sizing",
        "Generate the Surface Mesh",
        "Describe Geometry",
        "Apply Share Topology",
        "Enclose Fluid Regions (Capping)",
        "Update Boundaries",
        "Create Regions",
        "Update Regions",
        "Add Boundary Layers",
        "Generate the Volume Mesh",
    ]

    assert meshing.workflow.TaskObject.get_object_names() == child_object_names



import pprint
import pydoc
import sys


def docother(self, object, name=None, mod=None, parent=None, maxlen=None, doc=None):
    """Produce text documentation for a data object."""
    if isinstance(object, list):
        indent_len = len(name and name + " = " or "") + 1
        repr = pprint.pformat(object, width=maxlen, compact=True, indent=indent_len)
        repr = "[" + repr[1:].lstrip()
    else:
        repr = self.repr(object)
        if maxlen:
            line = (name and name + " = " or "") + repr
            chop = maxlen - len(line)
            if chop < 0:
                repr = repr[:chop] + "..."
    line = (name and self.bold(name) + " = " or "") + repr
    #  The source have been changed in 3.9, cpython commit id fbf2786c4c89430e2067016603078cf3500cfe94
    if sys.version_info < (3, 9):
        if doc is not None:
            line += "\n" + self.indent(str(doc))
    else:
        if not doc:
            doc = pydoc.getdoc(object)
        if doc:
            line += "\n" + self.indent(str(doc)) + "\n"
    return line



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from .file_name import file_name
from .state_name_1 import state_name
class write(Command):
    """
    Write display states to a file.
    
    Parameters
    ----------
        file_name : str
            'file_name' child.
        state_name : typing.List[str]
            'state_name' child.
    
    """

    fluent_name = "write"

    argument_names = \
        ['file_name', 'state_name']

    file_name: file_name = file_name
    """
    file_name argument of write.
    """
    state_name: state_name = state_name
    """
    state_name argument of write.
    """



from ansys.fluent.core.utils.fix_doc import fix_settings_doc


def test_fix_definition_list_in_settings_doc():
    old_doc = """Menu to define the rotor pitch and flapping angles.
 - blade-pitch-collective    : ,
 - blade-pitch-cyclic-sin    : ,
 - blade-pitch-cyclic-cos    : ,
For more details please consult the help option of the corresponding menu or TUI command."""
    new_doc = fix_settings_doc(old_doc)
    assert (
        new_doc
        == """Menu to define the rotor pitch and flapping angles.

 - blade-pitch-collective    : ,
 - blade-pitch-cyclic-sin    : ,
 - blade-pitch-cyclic-cos    : ,

For more details please consult the help option of the corresponding menu or TUI command."""
    )


def test_fix_wildcards_in_settings_doc():
    old_doc = """Read boundary profile data (*.prof, *.csv). Default is *.prof."""
    new_doc = fix_settings_doc(old_doc)
    assert (
        new_doc
        == r"""Read boundary profile data (\*.prof, \*.csv). Default is \*.prof."""
    )



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class two_stage_runge_kutta(Boolean):
    """
    'two_stage_runge_kutta' child.
    """

    fluent_name = "two-stage-runge-kutta?"




#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

class orig_beam_name(String):
    """
    Choose the name for the optical beam to be copied.
    """

    fluent_name = "orig-beam-name"

